from dataclasses import dataclass
from enum import Enum

import pandas as pd

from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    Measurement,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


class AggregatingProperty(Enum):
    ASSAY_IDENTIFIER = "assay identifier"
    SAMPLE_IDENTIFIER = "sample_identifier"
    WELL_PLATE_IDENTIFIER = "well_plate_identifier"
    LOCATION_IDENTIFIER = "location_identifier"
    WELL_PLATE_ID = "well_plate_identifier"
    LOCATION_ID = "location_identifier"


class CalculatedDataColumns(Enum):
    ADJUSTED_SIGNAL = "Adjusted Signal"
    MEAN = "Mean"
    ADJ_SIG_MEAN = "Adj. Sig. Mean"
    FIT_STATISTIC_RSQUARED = "Fit Statistic: RSquared"
    CV = "CV"
    PERCENT_RECOVERY = "% Recovery"
    PERCENT_RECOVERY_MEAN = "% Recovery Mean"
    CALC_CONCENTRATION = "Calc. Concentration"
    CALC_CONC_MEAN = "Calc. Conc. Mean"


@dataclass
class CalculatedDataMapping:
    msd_column: CalculatedDataColumns
    is_data_source_id_measurement: bool  # if True then data source id is measurement id, else calculated data id
    data_source_feature: str
    aggregating_property: list[str]


calculated_data_mappings = [
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.ADJUSTED_SIGNAL,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[
            AggregatingProperty.ASSAY_IDENTIFIER.value,
            AggregatingProperty.SAMPLE_IDENTIFIER.value,
        ],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.MEAN,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.ADJ_SIG_MEAN,
        is_data_source_id_measurement=False,
        data_source_feature=CalculatedDataColumns.MEAN.value,
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.FIT_STATISTIC_RSQUARED,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.CV,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.PERCENT_RECOVERY,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[
            AggregatingProperty.SAMPLE_IDENTIFIER.value,
            AggregatingProperty.WELL_PLATE_IDENTIFIER.value,
            AggregatingProperty.LOCATION_IDENTIFIER.value,
        ],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.PERCENT_RECOVERY_MEAN,
        is_data_source_id_measurement=False,
        data_source_feature=CalculatedDataColumns.PERCENT_RECOVERY.value,
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.CALC_CONCENTRATION,
        is_data_source_id_measurement=True,
        data_source_feature="luminescence",
        aggregating_property=[
            AggregatingProperty.SAMPLE_IDENTIFIER.value,
            AggregatingProperty.WELL_PLATE_IDENTIFIER.value,
            AggregatingProperty.LOCATION_IDENTIFIER.value,
        ],
    ),
    CalculatedDataMapping(
        msd_column=CalculatedDataColumns.CALC_CONC_MEAN,
        is_data_source_id_measurement=False,
        data_source_feature=CalculatedDataColumns.CALC_CONCENTRATION.value,
        aggregating_property=[AggregatingProperty.ASSAY_IDENTIFIER.value],
    ),
]


def _format_r_squared_name(name: str) -> str:
    return name.split(":")[-1].strip()


def _get_measurement_by_location_identifier(
    measurements: list[Measurement], well_location_identifier: str, spot_index: str
) -> Measurement:
    for measurement in measurements:
        if (
            measurement.well_location_identifier == well_location_identifier
            and measurement.location_identifier == spot_index
        ):
            return measurement
    msg = (
        f"No measurement found for well location identifier: {well_location_identifier}"
    )
    raise AllotropeConversionError(msg)


def _get_measurement_aggregate_properties(
    measurement: Measurement, aggregating_property: list[str]
) -> list[str]:
    aggregating_property = aggregating_property.copy()
    agg_properties: list[str] = []
    if (
        AggregatingProperty.ASSAY_IDENTIFIER.value in aggregating_property
        and measurement.measurement_custom_info
    ):
        assay_id = measurement.measurement_custom_info.get(
            AggregatingProperty.ASSAY_IDENTIFIER.value
        )
        if assay_id:
            agg_properties.append(assay_id)
        del aggregating_property[
            aggregating_property.index(AggregatingProperty.ASSAY_IDENTIFIER.value)
        ]

    agg_properties.extend([getattr(measurement, prop) for prop in aggregating_property])
    return agg_properties


def _get_data_sources(
    measurements: list[Measurement],
    calculated_data_mapping: CalculatedDataMapping,
    row: SeriesData,
    calculated_data: list[CalculatedDataItem],
) -> list[DataSource]:
    well_location_identifier = row[str, "Well"]
    spot_index = row[str, "Spot"]
    current_measurement = _get_measurement_by_location_identifier(
        measurements, well_location_identifier, spot_index
    )
    agg_properties = _get_measurement_aggregate_properties(
        current_measurement, calculated_data_mapping.aggregating_property
    )
    # get measurements with the same aggregate properties as the current one
    measurements_with_agg_properties = [
        measurement
        for measurement in measurements
        if _get_measurement_aggregate_properties(
            measurement, calculated_data_mapping.aggregating_property
        )
        == agg_properties
    ]
    if not calculated_data_mapping.is_data_source_id_measurement:
        feature = calculated_data_mapping.data_source_feature
        if feature == CalculatedDataColumns.FIT_STATISTIC_RSQUARED.value:
            feature = _format_r_squared_name(feature)
        calc_data_by_feature = list(
            filter(lambda item: item.name == feature, calculated_data)
        )
        measurements_with_agg_properties_ids = [
            measurement.identifier for measurement in measurements_with_agg_properties
        ]
        calc_data_by_feature = list(
            filter(
                lambda calc: all(
                    source.identifier in measurements_with_agg_properties_ids
                    for source in calc.data_sources
                ),
                calc_data_by_feature,
            )
        )
        return [
            DataSource(
                identifier=calc_data.identifier,
                feature=calculated_data_mapping.data_source_feature,
            )
            for calc_data in calc_data_by_feature
        ]

    return [
        DataSource(
            identifier=measurement.identifier,
            feature=calculated_data_mapping.data_source_feature,
        )
        for measurement in measurements_with_agg_properties
    ]


def _is_calc_data_created(
    calculated_data: list[CalculatedDataItem],
    data_sources: list[DataSource],
    calc_data_name: str,
) -> bool:
    for calc_data in calculated_data:
        if calc_data.name == calc_data_name and set(calc_data.data_sources) == set(
            data_sources
        ):
            return True
    return False


def create_calculated_data_groups(
    data: pd.DataFrame, measurements: list[Measurement]
) -> list[CalculatedDataItem]:
    calculated_data: list[CalculatedDataItem] = []
    for _row_index, row in data.iterrows():
        row_series = SeriesData(row)
        for calc_data_item in calculated_data_mappings:
            value = row_series.get(float, calc_data_item.msd_column.value)
            data_sources = _get_data_sources(
                measurements, calc_data_item, row_series, calculated_data
            )
            name = calc_data_item.msd_column.value
            # handle special case for Fit Statistic: RSquared
            if name == CalculatedDataColumns.FIT_STATISTIC_RSQUARED.value:
                name = _format_r_squared_name(name)
            if (
                not value
                or not data_sources
                or _is_calc_data_created(calculated_data, data_sources, name)
            ):
                continue

            calculated_data.append(
                CalculatedDataItem(
                    identifier=random_uuid_str(),
                    data_sources=data_sources,
                    unit=UNITLESS,
                    name=name,
                    value=value,
                )
            )

    # merge calculated data that has its source as other calculated data values
    not_data_source_measurement_id_mappings = [
        calc_data_mapping
        for calc_data_mapping in calculated_data_mappings
        if not calc_data_mapping.is_data_source_id_measurement
    ]
    for calc_data_mapping in not_data_source_measurement_id_mappings:
        calc_data: list[CalculatedDataItem] = [
            calc_data_item
            for calc_data_item in calculated_data
            if calc_data_item.name == calc_data_mapping.msd_column.value
        ]
        if not calc_data:
            continue
        data_sources_identifiers = [
            source.identifier for calc in calc_data for source in calc.data_sources
        ]
        if len(data_sources_identifiers) != len(set(data_sources_identifiers)):
            ordered_sources = list(set(data_sources_identifiers))
            ordered_sources.sort()
            value = calc_data[0].value
            new_calc_data_item = CalculatedDataItem(
                identifier=random_uuid_str(),
                data_sources=[
                    DataSource(
                        identifier=identifier,
                        feature=calc_data_mapping.data_source_feature,
                    )
                    for identifier in ordered_sources
                ],
                unit=UNITLESS,
                name=calc_data_mapping.msd_column.value,
                value=value,
            )
            for item in calc_data:
                calculated_data.remove(item)
            calculated_data.append(new_calc_data_item)
    return calculated_data



import pandas as pd

from allotropy.parsers.utils.pandas import assert_not_empty_df, read_csv
from allotropy.types import IOType


class NucleoviewReader:
    SUPPORTED_EXTENSIONS = "csv"

    @classmethod
    def read(cls, contents: IOType) -> pd.DataFrame:
        df = read_csv(contents, sep=";", skipinitialspace=True, index_col=0)
        df = df[:-1].dropna(axis="index", how="all").T
        df = df.rename(
            {"Estimated cell diameter [um]": "Estimated cell diameter (um)"},
            axis="columns",
        )
        # Drop rows that do not report cell viability.
        df = df.dropna(subset=["Viability (%)"])
        return assert_not_empty_df(df, "Unable to parse data from empty dataset.")



from collections.abc import Iterable
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    NaN,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import (
    CalculatedData,
    CalculatedDataItem,
    DataSource,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
    ProcessedData,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.appbio_quantstudio import constants
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
    Header,
    MeltCurveRawData,
    MulticomponentData,
    Result,
    Well,
    WellItem,
)
from allotropy.parsers.constants import get_well_count_by_well_ids, NEGATIVE_ZERO
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)


def _create_processed_data_cubes(
    amplification_data: AmplificationData | None,
) -> tuple[DataCube | None, DataCube | None]:
    if not amplification_data:
        return None, None
    cycle_count = DataCubeComponent(FieldComponentDatatype.double, "cycle count", "#")
    return (
        DataCube(
            label="normalized reporter",
            structure_dimensions=[cycle_count],
            structure_measures=[
                DataCubeComponent(
                    FieldComponentDatatype.double,
                    "normalized reporter result",
                    UNITLESS,
                )
            ],
            dimensions=[amplification_data.cycle],
            measures=[amplification_data.rn],
        ),
        DataCube(
            label="baseline corrected reporter",
            structure_dimensions=[cycle_count],
            structure_measures=[
                DataCubeComponent(
                    FieldComponentDatatype.double,
                    "baseline corrected reporter result",
                    UNITLESS,
                )
            ],
            dimensions=[amplification_data.cycle],
            measures=[amplification_data.delta_rn],
        ),
    )


def _create_processed_data(
    amplification_data: AmplificationData | None, result: Result
) -> ProcessedData:
    (
        normalized_reporter_data_cube,
        baseline_corrected_reporter_data_cube,
    ) = _create_processed_data_cubes(amplification_data)
    return ProcessedData(
        automatic_cycle_threshold_enabled_setting=result.automatic_cycle_threshold_enabled_setting,
        cycle_threshold_value_setting=result.cycle_threshold_value_setting,
        automatic_baseline_determination_enabled_setting=result.automatic_baseline,
        baseline_determination_start_cycle_setting=result.baseline_start,
        baseline_determination_end_cycle_setting=result.baseline_end,
        genotyping_determination_method_setting=result.genotyping_determination_method_setting,
        genotyping_determination_result=result.genotyping_determination_result,
        cycle_threshold_result=result.cycle_threshold_result,
        normalized_reporter_result=result.normalized_reporter_result,
        baseline_corrected_reporter_result=result.baseline_corrected_reporter_result,
        normalized_reporter_data_cube=normalized_reporter_data_cube,
        baseline_corrected_reporter_data_cube=baseline_corrected_reporter_data_cube,
        custom_info=result.extra_data,
    )


def _create_multicomponent_data_cubes(
    multicomponent_data: MulticomponentData | None,
    reporter_dye_setting: str | None,
    passive_reference_dye_setting: str | None,
) -> tuple[DataCube | None, DataCube | None]:
    if not multicomponent_data:
        return None, None

    cycle_count = DataCubeComponent(FieldComponentDatatype.double, "cycle count", "#")
    reporter_dye_data_cube = None
    if reporter_dye_setting is not None:
        reporter_dye_data_cube = DataCube(
            label="reporter dye",
            structure_dimensions=[cycle_count],
            structure_measures=[
                DataCubeComponent(
                    FieldComponentDatatype.double,
                    "fluorescence",
                    "RFU",
                )
            ],
            dimensions=[multicomponent_data.cycle],
            measures=[multicomponent_data.get_column(reporter_dye_setting)],
        )
    passive_reference_dye_data_cube = None
    if passive_reference_dye_setting is not None:
        passive_reference_dye_data_cube = DataCube(
            label="passive reference dye",
            structure_dimensions=[cycle_count],
            structure_measures=[
                DataCubeComponent(
                    FieldComponentDatatype.double,
                    "fluorescence",
                    "RFU",
                )
            ],
            dimensions=[multicomponent_data.cycle],
            measures=[multicomponent_data.get_column(passive_reference_dye_setting)],
        )
    return reporter_dye_data_cube, passive_reference_dye_data_cube


def _create_melt_curve_data_cube(
    melt_curve_raw_data: MeltCurveRawData | None,
) -> DataCube | None:
    if not melt_curve_raw_data:
        return None
    return DataCube(
        label="melting curve",
        structure_dimensions=[
            DataCubeComponent(FieldComponentDatatype.double, "temperature", "degC")
        ],
        structure_measures=[
            DataCubeComponent(
                FieldComponentDatatype.double,
                "fluorescence",
                "RFU",
            ),
            DataCubeComponent(FieldComponentDatatype.double, "slope", UNITLESS),
        ],
        dimensions=[melt_curve_raw_data.reading],
        measures=[
            melt_curve_raw_data.fluorescence,
            melt_curve_raw_data.derivative,
        ],
    )


def _create_measurement(
    well_item: WellItem,
    header: Header,
    multicomponent_data: MulticomponentData | None,
    melt_curve_raw_data: MeltCurveRawData | None,
    amplification_data: AmplificationData | None,
    result: Result | None,
) -> Measurement | None:
    if not result:
        return None
    # TODO: temp workaround for cal doc result
    well_item._result = result

    (
        reporter_dye_data_cube,
        passive_reference_dye_data_cube,
    ) = _create_multicomponent_data_cubes(
        multicomponent_data,
        well_item.reporter_dye_setting,
        header.passive_reference_dye_setting,
    )
    return Measurement(
        identifier=well_item.uuid,
        timestamp=header.measurement_time,
        target_identifier=well_item.target_dna_description,
        sample_identifier=well_item.sample_identifier,
        group_identifier=well_item.group_identifier,
        sample_role_type=well_item.sample_role_type,
        location_identifier=well_item.location_identifier,
        well_location_identifier=well_item.well_location_identifier,
        well_plate_identifier=header.barcode,
        total_cycle_number_setting=(
            amplification_data.total_cycle_number_setting
            if amplification_data
            else None
        ),
        pcr_detection_chemistry=header.pcr_detection_chemistry,
        reporter_dye_setting=well_item.reporter_dye_setting,
        quencher_dye_setting=well_item.quencher_dye_setting,
        passive_reference_dye_setting=header.passive_reference_dye_setting,
        processed_data=_create_processed_data(amplification_data, result),
        sample_custom_info=well_item.extra_data,
        reporter_dye_data_cube=reporter_dye_data_cube,
        passive_reference_dye_data_cube=passive_reference_dye_data_cube,
        melting_curve_data_cube=_create_melt_curve_data_cube(melt_curve_raw_data),
    )


def create_metadata(header: Header, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        device_identifier=header.device_identifier,
        device_type=constants.DEVICE_TYPE,
        device_serial_number=header.device_serial_number,
        model_number=header.model_number,
        software_name=constants.SOFTWARE_NAME,
        software_version=constants.SOFTWARE_VERSION,
        data_system_instance_identifier=constants.DATA_SYSTEM_INSTANCE_IDENTIFIER,
        file_name=path.name,
        unc_path=file_path,
        measurement_method_identifier=header.measurement_method_identifier,
        experiment_type=header.experiment_type.value,
        container_type=constants.CONTAINER_TYPE,
    )


def create_calculated_data(
    calculated_data_documents: Iterable[CalculatedDocument],
) -> CalculatedData:
    return CalculatedData(
        items=[
            CalculatedDataItem(
                identifier=cal_doc.uuid,
                name=cal_doc.name,
                value=cal_doc.value,
                unit=UNITLESS,
                data_sources=[
                    DataSource(
                        identifier=data_source.reference.uuid,
                        feature=data_source.feature,
                    )
                    for data_source in cal_doc.data_sources
                ],
            )
            for cal_doc in calculated_data_documents
        ],
    )


def get_well_item_results(
    well_item: WellItem,
    results_data: dict[int, dict[str, Result]],
) -> Result | None:
    return results_data.get(well_item.identifier, {}).get(
        well_item.target_dna_description.replace(" ", "")
    )


def get_well_item_amp_data(
    well_item: WellItem, amp_data: dict[int, dict[str, AmplificationData]]
) -> AmplificationData | None:
    return amp_data.get(well_item.identifier, {}).get(well_item.target_dna_description)


def _create_measurement_group(
    header: Header,
    well: Well,
    amp_data: dict[int, dict[str, AmplificationData]],
    multi_data: dict[int, MulticomponentData],
    results_data: dict[int, dict[str, Result]],
    melt_data: dict[int, MeltCurveRawData],
    plate_well_count: int | None,
) -> MeasurementGroup | None:
    measurements = [
        _create_measurement(
            well_item,
            header,
            multi_data.get(well.identifier),
            melt_data.get(well.identifier),
            get_well_item_amp_data(well_item, amp_data),
            get_well_item_results(well_item, results_data),
        )
        for well_item in well.items
        if get_well_item_results(well_item, results_data)
    ]
    group = MeasurementGroup(
        analyst=header.analyst,
        experimental_data_identifier=header.experimental_data_identifier,
        plate_well_count=plate_well_count,
        well_volume=header.well_volume,
        error_document=(
            [Error(error=NaN.value, feature="well volume")]
            if header.well_volume == NEGATIVE_ZERO
            else []
        ),
        measurements=[m for m in measurements if m is not None],
    )
    return group if group.measurements else None


def _get_plate_well_count(header: Header, wells: list[Well]) -> int | None:
    if header.plate_well_count is not None:
        return header.plate_well_count

    return get_well_count_by_well_ids(
        well_identifiers=[
            well_item.identifier for well in wells for well_item in well.items
        ],
        well_locations=[
            well_item.position
            for well in wells
            for well_item in well.items
            if well_item.position
        ],
    )


def create_measurement_groups(
    header: Header,
    wells: list[Well],
    amp_data: dict[int, dict[str, AmplificationData]],
    multi_data: dict[int, MulticomponentData],
    results_data: dict[int, dict[str, Result]],
    melt_data: dict[int, MeltCurveRawData],
) -> list[MeasurementGroup]:

    plate_well_count = _get_plate_well_count(header, wells)
    groups = [
        _create_measurement_group(
            header,
            well,
            amp_data,
            multi_data,
            results_data,
            melt_data,
            plate_well_count,
        )
        for well in wells
    ]
    return [group for group in groups if group]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_visionlite.constants import (
    DETECTION_TYPE,
    DEVICE_TYPE,
    PRODUCT_MANUFACTURER,
    SAMPLE_NAME_COLS,
    SOFTWARE_NAME,
    UNSUPPORTED_KINETIC_MEASUREMENTS_ERROR,
)
from allotropy.parsers.thermo_fisher_visionlite.thermo_fisher_visionlite_reader import (
    ThermoFisherVisionliteReader,
)
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_none,
    try_int_or_none,
)


class ExperimentType(Enum):
    SCAN = "Scan"
    QUANT = "Quant"
    FIXED = "Fixed"

    @property
    def measurement_type(self) -> MeasurementType:
        return _EXPERIMENT_TO_MEASUREMENT_TYPE[self]


_EXPERIMENT_TO_MEASUREMENT_TYPE = {
    ExperimentType.FIXED: MeasurementType.ULTRAVIOLET_ABSORBANCE,
    ExperimentType.QUANT: MeasurementType.ULTRAVIOLET_ABSORBANCE,
    ExperimentType.SCAN: MeasurementType.ULTRAVIOLET_ABSORBANCE_SPECTRUM,
}


def _get_experiment_type(reader: ThermoFisherVisionliteReader) -> ExperimentType:
    if reader.header:
        if "nm" in reader.data.columns:
            return ExperimentType.SCAN
        else:
            raise AllotropeConversionError(UNSUPPORTED_KINETIC_MEASUREMENTS_ERROR)
    elif "Dilution factor" in reader.data.columns:
        return ExperimentType.QUANT
    else:
        return ExperimentType.FIXED


@dataclass(frozen=True)
class Header:
    sample_name: str | None = None
    analyst: str | None = None
    measurement_time: str = DEFAULT_EPOCH_TIMESTAMP

    @staticmethod
    def create(header_data: SeriesData | None) -> Header:
        if header_data is None:
            return Header()

        return Header(
            sample_name=header_data[str, ("Sample Name", "Well Position")],
            analyst=header_data[str, "Analyst"],
            measurement_time=header_data[str, "Measurement Time"],
        )


@dataclass(frozen=True)
class AbsorbanceMeasurement:
    absorbance: float
    wavelength: int | None = None


class VisionLiteData(Data):
    @staticmethod
    def create(
        reader: ThermoFisherVisionliteReader,
        file_path: str,
        software_name: str = SOFTWARE_NAME,
    ) -> Data:
        experiment_type = _get_experiment_type(reader)
        header = Header.create(reader.header)
        data = reader.data
        measurement_groups = []
        calculated_data = []

        if experiment_type == ExperimentType.SCAN:
            if header.sample_name is None:
                msg = "Missing Sample Name."
                raise AllotropeConversionError(msg)
            measurement_groups = [
                MeasurementGroup(
                    analyst=header.analyst,
                    measurement_time=header.measurement_time,
                    experiment_type=experiment_type.value,
                    measurements=[
                        Measurement(
                            type_=experiment_type.measurement_type,
                            identifier=random_uuid_str(),
                            sample_identifier=header.sample_name,
                            data_cube=_get_data_cube(data),
                        )
                    ],
                )
            ]
        else:
            wavelength_cols = _get_wavelength_cols(list(data.columns))
            for _, row in data.iterrows():
                row_data = SeriesData(row)
                measurements = _get_absorbance_measurements(
                    row_data, experiment_type, wavelength_cols
                )
                measurement_groups.append(
                    MeasurementGroup(
                        analyst=header.analyst,
                        measurement_time=header.measurement_time,
                        experiment_type=experiment_type.value,
                        measurements=measurements,
                    )
                )
                if result := row_data.get(float, "Result"):
                    calculated_data.append(
                        _get_calculated_data_item(result, measurements)
                    )

        return Data(
            metadata=_create_metadata(file_path, software_name),
            measurement_groups=measurement_groups,
            calculated_data=calculated_data,
        )


def _get_calculated_data_item(
    result: float, measurements: list[Measurement]
) -> CalculatedDataItem:
    return CalculatedDataItem(
        identifier=random_uuid_str(),
        name="Result",
        value=result,
        unit=UNITLESS,
        data_sources=[
            DataSource(identifier=measurement.identifier, feature="Absorbance")
            for measurement in measurements
        ],
    )


def _create_metadata(file_path: str, software_name: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        device_identifier=NOT_APPLICABLE,
        device_type=DEVICE_TYPE,
        model_number=NOT_APPLICABLE,
        software_name=software_name,
        detection_type=DETECTION_TYPE,
        product_manufacturer=PRODUCT_MANUFACTURER,
    )


def _get_absorbance_measurements(
    data: SeriesData,
    experiment_type: ExperimentType,
    wavelength_cols: dict[int, str],
) -> list[Measurement]:
    if experiment_type == ExperimentType.QUANT:
        ordinate_col = "Ordinate [A]"
        if not data.has_key(ordinate_col):
            msg = "Unable to determine Quant absorbance measurements, Ordinate column missing."
            raise AllotropeConversionError(msg)
        absorbance_measurements = [
            AbsorbanceMeasurement(absorbance=data[float, ordinate_col])
        ]
    elif experiment_type == ExperimentType.FIXED:
        if wavelength_cols:
            absorbance_measurements = [
                AbsorbanceMeasurement(
                    absorbance=data[float, col],
                    wavelength=wavelength,
                )
                for wavelength, col in wavelength_cols.items()
            ]
        else:
            absorbance_measurements = [
                AbsorbanceMeasurement(absorbance=data[float, "Absorbance"])
            ]

    return [
        Measurement(
            type_=experiment_type.measurement_type,
            identifier=random_uuid_str(),
            sample_identifier=data[str, SAMPLE_NAME_COLS],
            processed_data=(
                _get_processed_data(data)
                if experiment_type == ExperimentType.QUANT
                else None
            ),
            absorbance=measurement.absorbance,
            dilution_factor_setting=data.get(float, "Dilution factor"),
            detector_wavelength_setting=try_float_or_none(measurement.wavelength),
        )
        for measurement in absorbance_measurements
    ]


def _get_processed_data(data: SeriesData) -> ProcessedData:
    concentration_unit = _get_concentration_unit(list(data.series.index))

    concentration_col = f"Concentration [{concentration_unit}]"
    if not data.has_key(concentration_col):
        msg = "Unable to find Concentration data."
        raise AllotropeConversionError(msg)

    return ProcessedData(
        identifier=random_uuid_str(),
        features=[
            ProcessedDataFeature(
                result=data[float, concentration_col],
                unit=concentration_unit,
            )
        ],
    )


def _get_concentration_unit(columns: list[str]) -> str:
    for col in columns:
        if match := re.match(r"Concentration \[(.*)\]", col):
            return match.groups()[0]
    return ""


def _get_wavelength_cols(columns: list[str]) -> dict[int, str]:
    wavelength_cols = {}
    for col in columns:
        if match := re.match(r"(\d{1,}) nm   \[A\]|A(\d{1,})", col):
            wavelength = assert_not_none(
                get_first_not_none(try_int_or_none, match.groups())
            )
            wavelength_cols[wavelength] = col
    return wavelength_cols


def _get_data_cube(data: pd.DataFrame) -> DataCube:
    return DataCube(
        label="absorption spectrum",
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="wavelength",
                unit="nm",
            )
        ],
        structure_measures=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="absorbance",
                unit="mAU",
            )
        ],
        dimensions=[data["nm"].to_list()],
        measures=[data["A"].to_list()],
    )



from allotropy.allotrope.models.adm.binding_affinity_analyzer.wd._2024._12.binding_affinity_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cytiva_biacore_t200_control import constants
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_decoder import (
    decode_data,
)
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CytivaBiacoreT200ControlParser(VendorParser[Data, Model]):
    DISPLAY_NAME = constants.DISPLAY_NAME
    RELEASE_STATE = ReleaseState.WORKING_DRAFT
    SUPPORTED_EXTENSIONS = "blr"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        structured_data = decode_data(named_file_contents)
        return Data(
            metadata=create_metadata(structured_data, named_file_contents),
            measurement_groups=[create_measurement_groups(structured_data)],
        )



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import PureWindowsPath
import re

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import SampleRoleType
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
    MulticomponentData,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.constants import (
    ExperimentType,
)
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    Referenceable,
)
from allotropy.parsers.utils.pandas import (
    assert_df_column,
    assert_not_empty_df,
    SeriesData,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_int,
)

SAMPLE_ROLE_TYPES_MAP = {
    "NTC": SampleRoleType.control_sample_role,
    "STANDARD": SampleRoleType.standard_sample_role,
    "UNKNOWN": SampleRoleType.unknown_sample_role,
    "POSITIVE_CONTROL": SampleRoleType.control_sample_role,
    "IPC": SampleRoleType.control_sample_role,
    "BLOCKED_IPC": SampleRoleType.control_sample_role,
    "BLOCKEDIPC": SampleRoleType.control_sample_role,
    "PC_ALLELE_1": SampleRoleType.control_sample_role,
    "PC_ALLELE_2": SampleRoleType.control_sample_role,
    "PC_ALLELE_BOTH": SampleRoleType.control_sample_role,
    "POSITIVE_11": SampleRoleType.control_sample_role,
    "POSITIVE_22": SampleRoleType.control_sample_role,
    "POSITIVE_12": SampleRoleType.control_sample_role,
    "Accuracy Control": SampleRoleType.control_sample_role,
}


def get_well_volume(block_type: str) -> float:
    # if the block type includes the well volume in mL, convert to microliters
    if well_search := re.search(r"([0-9]+\.[0-9]+)-?mL", block_type):
        return float(well_search.groups()[0]) * 1000
    # The 384-well block and Taqman Array Card have specified well volumes from the
    # manufacturer (Thermo Fisher) which are 40uL and 1.5uL, respectively.
    elif "384-Well Block" in block_type:
        return 40
    elif "Taqman Array Card" in block_type:
        return 1.5
    # Since well volume is required, if it cannot be implied from block type
    # a negative zero will be returned, indicating an error.
    return NEGATIVE_ZERO


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    well_volume: float
    pcr_detection_chemistry: str | None
    passive_reference_dye_setting: str | None
    barcode: str | None
    analyst: str | None
    experimental_data_identifier: str
    pcr_stage_number: int | None
    software_name: str | None
    software_version: str | None
    block_serial_number: str | None
    heated_cover_serial_number: str | None

    @staticmethod
    def create(header: SeriesData) -> Header:
        software_info = re.match(
            "(.*) v(.+)",
            header.get(str, "Software Name and Version", ""),
        )

        software_name, software_version = (
            (software_info.group(1), software_info.group(2))
            if software_info
            else (None, None)
        )

        stage_number_raw = header.get(str, "PCR Stage/Step Number", "")
        stage_number = re.match(r"Stage (\d+)", stage_number_raw)
        pcr_stage_number = None if stage_number is None else int(stage_number.group(1))
        block_type = header[str, "Block Type"]

        return Header(
            measurement_time=header[
                str,
                ["Run End Date/Time", "Run End Data/Time"],
                "Unable to find measurement time.",
            ],
            plate_well_count=assert_not_none(
                try_int(
                    assert_not_none(
                        re.match("(96)|(384)", block_type),
                        msg="Unable to find plate well count",
                    ).group(),
                    "plate well count",
                ),
                msg="Unable to interpret plate well count",
            ),
            device_identifier=header.get(str, "Instrument Name", NOT_APPLICABLE),
            model_number=header.get(str, "Instrument Type", NOT_APPLICABLE),
            device_serial_number=header.get(
                str, "Instrument Serial Number", NOT_APPLICABLE
            ),
            measurement_method_identifier=header[str, "Quantification Cycle Method"],
            pcr_detection_chemistry=header.get(str, "Chemistry"),
            passive_reference_dye_setting=header.get(str, "Passive Reference"),
            barcode=header.get(str, "Barcode"),
            analyst=header.get(str, "Operator"),
            experimental_data_identifier=PureWindowsPath(header[str, "File Name"]).name,
            block_serial_number=header.get(str, "Block Serial Number"),
            heated_cover_serial_number=header.get(str, "Heated Cover Serial Number"),
            pcr_stage_number=pcr_stage_number,
            software_name=software_name,
            software_version=software_version,
            well_volume=get_well_volume(block_type),
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    location_identifier: str
    reporter_dye_setting: str | None
    well_location_identifier: str | None
    quencher_dye_setting: str | None
    sample_role_type: SampleRoleType | None
    amplification_data: AmplificationData | None
    result: Result
    melt_curve_data: MeltCurveData | None = None

    @classmethod
    def get_result_class(cls) -> type[Result]:
        return Result

    # Make hashable to allow for use of caching
    def __hash__(self) -> int:
        return hash(self.identifier)

    @classmethod
    def get_amplification_data_sheet(
        cls, reader: DesignQuantstudioReader
    ) -> pd.DataFrame | None:
        return reader.get_non_empty_sheet("Amplification Data")

    @classmethod
    def create(
        cls,
        reader: DesignQuantstudioReader,
        data: SeriesData,
    ) -> WellItem:
        identifier = data[int, "Well"]

        target_dna_description = data[
            str,
            "Target",
            f"Unable to find target dna description for well {identifier}",
        ]
        well_position = data[
            str,
            "Well Position",
            f"Unable to find well position for Well '{identifier}'.",
        ]

        amp_data = cls.get_amplification_data_sheet(reader)
        amplification_data = (
            create_amplification_data(amp_data, identifier, target_dna_description)
            if amp_data is not None
            else None
        )

        melt_data = reader.get_non_empty_sheet_or_none("Melt Curve Raw")
        melt_curve_data = (
            MeltCurveData.create(melt_data, identifier, target_dna_description)
            if melt_data is not None
            else None
        )

        result_class = cls.get_result_class()
        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=target_dna_description,
            sample_identifier=data.get(str, "Sample", well_position),
            reporter_dye_setting=data.get(str, "Reporter"),
            well_location_identifier=well_position,
            location_identifier=str(identifier),
            quencher_dye_setting=data.get(str, "Quencher"),
            sample_role_type=SAMPLE_ROLE_TYPES_MAP.get(
                data.get(str, "Task", "__INVALID_KEY__")
            ),
            amplification_data=amplification_data,
            melt_curve_data=melt_curve_data,
            result=result_class.create(data, identifier),
        )


@dataclass
class Well:
    identifier: int
    items: dict[str, WellItem]
    multicomponent_data: MulticomponentData | None = None

    def get_well_item(self, target: str) -> WellItem:
        return assert_not_none(
            self.items.get(target),
            msg=f"Unable to find target DNA '{target}' for well {self.identifier}.",
        )

    @classmethod
    def get_well_item_class(cls) -> type[WellItem]:
        return WellItem

    @classmethod
    def create(
        cls,
        reader: DesignQuantstudioReader,
        header: Header,
        well_data: pd.DataFrame,
        identifier: int,
    ) -> Well:
        well_item_class = cls.get_well_item_class()
        well_items = {
            SeriesData(item_data)[str, "Target"]: well_item_class.create(
                reader,
                SeriesData(item_data),
            )
            for _, item_data in well_data.iterrows()
        }
        multi_data = reader.get_non_empty_sheet_or_none("Multicomponent")
        return Well(
            identifier=identifier,
            items=well_items,
            multicomponent_data=(
                None
                if multi_data is None
                else create_multicomponent_data(header, multi_data, identifier)
            ),
        )


@dataclass(frozen=True)
class WellList:
    wells: list[Well]

    def get_well_items(self) -> list[WellItem]:
        wells: list[WellItem] = []
        for well in self.wells:
            wells += well.items.values()
        return wells

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)

    @classmethod
    def get_well_class(cls) -> type[Well]:
        return Well

    @classmethod
    def get_data_sheet(cls) -> str:
        return "Results"

    @classmethod
    def _add_data(
        cls, data: pd.DataFrame, extra_data: pd.DataFrame, columns: list[str]
    ) -> pd.DataFrame:
        new_data = data.copy()
        new_data[columns] = None
        for _, row in extra_data.iterrows():
            sample_cond = new_data["Sample"] == row["Sample"]
            target_cond = new_data["Target"] == row["Target"]
            new_data.loc[sample_cond & target_cond, columns] = row[columns].to_list()
        return new_data

    @classmethod
    def get_well_result_data(cls, reader: DesignQuantstudioReader) -> pd.DataFrame:
        return reader.get_non_empty_sheet(cls.get_data_sheet())

    @classmethod
    def create(
        cls,
        reader: DesignQuantstudioReader,
        header: Header,
    ) -> WellList:
        results_data = cls.get_well_result_data(reader)
        assert_df_column(results_data, "Well")

        well_class = cls.get_well_class()
        return WellList(
            wells=[
                well_class.create(
                    reader,
                    header,
                    well_data,
                    try_int(str(identifier), "well identifier"),
                )
                for identifier, well_data in results_data.groupby("Well")
            ]
        )


def create_amplification_data(
    amplification_data: pd.DataFrame,
    well_item_id: int,
    target_dna_description: str,
) -> AmplificationData:
    well_data = assert_not_empty_df(
        amplification_data[
            assert_df_column(amplification_data, "Well") == well_item_id
        ],
        msg=f"Unable to find amplification data for well {well_item_id}.",
    )

    target_data = assert_not_empty_df(
        well_data[assert_df_column(well_data, "Target") == target_dna_description],
        msg=f"Unable to find amplification data for target '{target_dna_description}' in well {well_item_id}.",
    )

    cycle_number = assert_df_column(target_data, "Cycle Number")
    return AmplificationData(
        total_cycle_number_setting=try_float(str(cycle_number.max()), "Cycle Number"),
        cycle=cycle_number.tolist(),
        rn=assert_df_column(target_data, "Rn").tolist(),
        delta_rn=assert_df_column(target_data, "dRn").tolist(),
    )


def create_multicomponent_data(
    header: Header, data: pd.DataFrame, well_id: int
) -> MulticomponentData:
    well_data = assert_not_empty_df(
        data[assert_df_column(data, "Well") == well_id],
        msg=f"Unable to find multi component data for well {well_id}.",
    )

    stage_number = well_data.get("Stage Number")
    stage_data = (
        well_data
        if header.pcr_stage_number is None
        or stage_number is None
        or well_data["Stage Number"].isnull().all()
        else assert_not_empty_df(
            well_data[stage_number == header.pcr_stage_number],  # type: ignore[arg-type]
            msg=f"Unable to find multi component data for stage {header.pcr_stage_number}.",
        )
    )

    return MulticomponentData(
        cycle=assert_df_column(stage_data, "Cycle Number").tolist(),
        columns={
            str(name): stage_data[name].tolist()
            for name in stage_data
            if name
            not in [
                "Well",
                "Cycle Number",
                "Well Position",
                "Stage Number",
                "Step Number",
            ]
        },
    )


@dataclass(frozen=True)
class MeltCurveData:
    target: str
    temperature: list[float]
    fluorescence: list[float | None]
    derivative: list[float | None]

    @staticmethod
    def create(
        data: pd.DataFrame, well_id: int, target_dna_description: str
    ) -> MeltCurveData:
        well_data = assert_not_empty_df(
            data[assert_df_column(data, "Well") == well_id],
            msg=f"Unable to find melt curve data for well {well_id}.",
        )

        target_data = assert_not_empty_df(
            well_data[assert_df_column(well_data, "Target") == target_dna_description],
            msg=f"Unable to find melt curve data for target '{target_dna_description}' in well {well_id} .",
        )

        return MeltCurveData(
            target=target_dna_description,
            temperature=assert_df_column(target_data, "Temperature").tolist(),
            fluorescence=assert_df_column(target_data, "Fluorescence").tolist(),
            derivative=assert_df_column(target_data, "Derivative").tolist(),
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: float | None
    automatic_cycle_threshold_enabled_setting: bool | None
    automatic_baseline_determination_enabled_setting: bool | None
    normalized_reporter_result: float | None
    baseline_corrected_reporter_result: float | None
    baseline_determination_start_cycle_setting: int | None
    baseline_determination_end_cycle_setting: int | None
    genotyping_determination_result: str | None
    genotyping_determination_method_setting: float | None
    quantity: float | None
    quantity_mean: float | None
    quantity_sd: float | None
    ct_mean: float | None
    eq_ct_mean: float | None
    adj_eq_ct_mean: float | None
    ct_sd: float | None
    ct_se: float | None
    delta_ct_mean: float | None
    delta_ct_se: float | None
    delta_ct_sd: float | None
    delta_delta_ct: float | None
    rq: float | None
    rq_min: float | None
    rq_max: float | None
    rn_mean: float | None
    rn_sd: float | None
    y_intercept: float | None
    r_squared: float | None
    slope: float | None
    efficiency: float | None
    amp_score: float | None
    cq_conf: float | None

    @classmethod
    def get_genotyping_determination_result(cls, _: SeriesData) -> str | None:
        return None

    @classmethod
    def get_genotyping_determination_method_setting(cls, _: SeriesData) -> float | None:
        return None

    @staticmethod
    def get_reference_sample(reader: DesignQuantstudioReader) -> str:
        data = reader.get_non_empty_sheet("RQ Replicate Group Result")
        reference_data = data[assert_df_column(data, "Rq") == 1]
        reference_sample_array = assert_df_column(reference_data, "Sample").unique()

        if reference_sample_array.size != 1:
            msg = "Unable to infer reference sample, expecting a single row in sheet 'RQ Replicate Group Result' to have Rq == 1."
            raise AllotropeConversionError(msg)

        return str(reference_sample_array[0])

    @staticmethod
    def get_reference_target(reader: DesignQuantstudioReader) -> str | None:
        data = reader.get_non_empty_sheet("RQ Replicate Group Result")

        possible_ref_targets = set.intersection(
            *[
                set(
                    assert_df_column(
                        sample_data[assert_df_column(sample_data, "Rq").isnull()],
                        "Target",
                    ).tolist()
                )
                for _, sample_data in data.groupby("Sample")
            ]
        )

        if len(possible_ref_targets) == 0:
            return None

        if len(possible_ref_targets) == 1:
            return str(possible_ref_targets.pop())

        msg = "Unable to infer reference target, expecting a single unique value for Target in sheet 'RQ Replicate Group Result' where Rq is empty."
        raise AllotropeConversionError(msg)

    @classmethod
    def create(
        cls,
        target_data: SeriesData,
        well_item_id: int,
    ) -> Result:
        return Result(
            cycle_threshold_value_setting=target_data[
                float,
                "Threshold",
                f"Unable to find cycle threshold value setting for well {well_item_id}",
            ],
            cycle_threshold_result=target_data.get(float, "Cq"),
            automatic_cycle_threshold_enabled_setting=target_data.get(
                bool, "Auto Threshold", None
            ),
            automatic_baseline_determination_enabled_setting=target_data.get(
                bool, "Auto Baseline", None
            ),
            normalized_reporter_result=target_data.get(float, "Rn"),
            baseline_corrected_reporter_result=target_data.get(float, "Delta Rn"),
            baseline_determination_start_cycle_setting=target_data.get(
                int, "Baseline Start"
            ),
            baseline_determination_end_cycle_setting=target_data.get(
                int, "Baseline End"
            ),
            genotyping_determination_result=cls.get_genotyping_determination_result(
                target_data
            ),
            genotyping_determination_method_setting=cls.get_genotyping_determination_method_setting(
                target_data
            ),
            quantity=target_data.get(float, "Quantity"),
            quantity_mean=target_data.get(float, "Quantity Mean"),
            quantity_sd=target_data.get(float, "Quantity SD"),
            ct_mean=target_data.get(float, "Cq Mean"),
            eq_ct_mean=target_data.get(float, "EqCq Mean"),
            adj_eq_ct_mean=target_data.get(float, "Adjusted EqCq Mean"),
            ct_sd=target_data.get(float, "Cq SD"),
            ct_se=target_data.get(float, "Cq SE"),
            delta_ct_mean=target_data.get(float, "Delta EqCq Mean"),
            delta_ct_se=target_data.get(float, "Delta EqCq SE"),
            delta_ct_sd=target_data.get(float, "Delta EqCq SD"),
            delta_delta_ct=target_data.get(float, "Delta Delta EqCq"),
            rq=target_data.get(float, "Rq"),
            rq_min=target_data.get(float, "Rq Min"),
            rq_max=target_data.get(float, "Rq Max"),
            rn_mean=target_data.get(float, "Rn Mean"),
            rn_sd=target_data.get(float, "Rn SD"),
            y_intercept=target_data.get(float, "Y-Intercept"),
            r_squared=target_data.get(float, "R2"),
            slope=target_data.get(float, "Slope"),
            efficiency=target_data.get(float, "Efficiency"),
            amp_score=target_data.get(float, "Amp Score"),
            cq_conf=target_data.get(float, "Cq Confidence"),
        )


@dataclass(frozen=True)
class Data:
    header: Header
    wells: WellList
    experiment_type: ExperimentType
    calculated_documents: list[CalculatedDocument]
    reference_target: str | None
    reference_sample: str | None



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_genesys_on_board.constants import (
    DISPLAY_NAME,
    SOFTWARE_NAME,
)
from allotropy.parsers.thermo_fisher_visionlite.thermo_fisher_visionlite_reader import (
    ThermoFisherVisionliteReader,
)
from allotropy.parsers.thermo_fisher_visionlite.thermo_fisher_visionlite_structure import (
    VisionLiteData,
)
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherGenesysOnBoardParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = ThermoFisherVisionliteReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        return VisionLiteData.create(
            ThermoFisherVisionliteReader(named_file_contents),
            named_file_contents.original_file_path,
            software_name=SOFTWARE_NAME,
        )



from __future__ import annotations

from dataclasses import dataclass

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass
class StaticDocs:
    chromatography_serial_num: str | None
    column_inner_diameter: float | None
    chromatography_chemistry_type: str | None
    chromatography_particle_size: float | None
    injection_identifier: str
    injection_time: str
    autosampler_injection_volume_setting: float
    sample_volume_2: float | None
    sample_volume_3: float | None
    sample_identifier: str
    sample_identifier_2: str | None
    sample_identifier_3: str | None
    batch_identifier: str | None
    start_time: str | None

    @classmethod
    def create(
        cls,
        handler: UnicornZipHandler,
        curve: StrictXmlElement,
        results: StrictXmlElement,
    ) -> StaticDocs:
        column_type_data = handler.get_column_type_data()
        autosampler_injection_volume_setting = NEGATIVE_ZERO
        try:
            inj_result = cls.__filter_result_criteria(results, keyword="Sample volume")
            autosampler_injection_volume_setting = inj_result.find(
                "Keyword2"
            ).get_float("Sample volume")
        except AllotropeConversionError:
            pass

        sample_volume_2 = None
        try:
            inj_result = cls.__filter_result_criteria(
                results, keyword="Sample volume_2"
            )
            sample_volume_2 = inj_result.find("Keyword2").get_float("sample_volume_2")
        except AllotropeConversionError:
            pass
        sample_volume_3 = None
        try:
            inj_result = cls.__filter_result_criteria(
                results, keyword="Sample volume_3"
            )
            sample_volume_3 = inj_result.find("Keyword2").get_float("Sample volume_3")
        except AllotropeConversionError:
            pass

        sample_identifier = NOT_APPLICABLE
        try:
            sample_result = cls.__filter_result_criteria(results, keyword="Sample_ID")
            sample_identifier = sample_result.find("Keyword2").get_text(
                "Sample_ID keyword2"
            )
        except AllotropeConversionError:
            pass

        sample_identifier_2 = None
        try:
            sample_result = cls.__filter_result_criteria(results, keyword="Sample_ID_2")
            sample_identifier = sample_result.find("Keyword2").get_text(
                "Sample_ID_2 keyword2"
            )
        except AllotropeConversionError:
            pass

        sample_identifier_3 = None
        try:
            sample_result = cls.__filter_result_criteria(results, keyword="Sample_ID_3")
            sample_identifier = sample_result.find("Keyword2").get_text(
                "Sample_ID_3 keyword2"
            )
        except AllotropeConversionError:
            pass

        article_number = column_type_data.recursive_find_or_none(
            ["ColumnType", "Hardware", "ArticleNumber"]
        )

        diameter = column_type_data.recursive_find_or_none(
            ["ColumnType", "Hardware", "Diameter"]
        )

        technique_name = column_type_data.recursive_find_or_none(
            ["ColumnType", "Media", "TechniqueName"]
        )

        avg_particle_diameter = column_type_data.recursive_find_or_none(
            ["ColumnType", "Media", "AverageParticleDiameter"]
        )

        batch_id = results.find_or_none("BatchId")

        return StaticDocs(
            chromatography_serial_num=(
                article_number.get_text_or_none()
                if article_number is not None
                else None
            ),
            column_inner_diameter=(
                diameter.get_float("column inner diameter") * 10
                if diameter is not None
                else None
            ),
            chromatography_chemistry_type=(
                technique_name.get_text_or_none()
                if technique_name is not None
                else None
            ),
            chromatography_particle_size=(
                avg_particle_diameter.get_float("chromatography particle size")
                if avg_particle_diameter is not None
                else None
            ),
            injection_identifier=random_uuid_str(),
            injection_time=curve.find("MethodStartTime").get_text("MethodStartTime"),
            autosampler_injection_volume_setting=autosampler_injection_volume_setting,
            sample_volume_2=sample_volume_2,
            sample_volume_3=sample_volume_3,
            sample_identifier=sample_identifier,
            sample_identifier_2=sample_identifier_2,
            sample_identifier_3=sample_identifier_3,
            batch_identifier=(
                batch_id.get_text_or_none() if batch_id is not None else None
            ),
            start_time=curve.get_sub_text_or_none("MethodStartTime"),
        )

    @classmethod
    def __filter_result_criteria(
        cls, results: StrictXmlElement, keyword: str
    ) -> StrictXmlElement:
        for result_criteria in results.find("ResultSearchCriterias").findall(
            "ResultSearchCriteria"
        ):
            if result_criteria.find("Keyword1").get_text_or_none() == keyword:
                return result_criteria
        msg = f"Unable to find result criteria with keyword 1 '{keyword}'"
        raise AllotropeConversionError(msg)



from collections.abc import Sequence
from enum import Enum

from attr import dataclass

PLATE_WELL_COUNT = 16
DEVICE_TYPE = "dPCR"
PRODUCT_MANUFACTURER = "ThermoFisher Scientific"
BRAND_NAME = "QuantStudio Absolute Q Digital PCR System"
SOFTWARE_NAME = "QuantStudio Absolute Q Digital PCR Software"
CONCENTRATION_COLUMNS = ("Conc. cp/uL", "Copies per microliter", "Conc.")

POSSIBLE_DYE_SETTING_LENGTHS = [3]


def get_dye_settings(columns: list[str]) -> list[str]:
    return [
        col
        for col in columns
        if len(col) in POSSIBLE_DYE_SETTING_LENGTHS and col == col.upper()
    ]


class AggregationType(str, Enum):
    INDIVIDUAL = "INDIVIDUAL"
    AVERAGE = "AVERAGE"
    POOLED = "POOLED"


class CalculatedDataSource(str, Enum):
    MEASUREMENT = "MEASUREMENT"
    CALCULATED_DATA = "CALCULATED_DATA"


AGGREGATION_LOOKUP = {
    "Average": AggregationType.AVERAGE,
    "Pooled": AggregationType.POOLED,
}


@dataclass(frozen=True)
class CalculatedDataReference:
    column: str | Sequence[str]
    name: str
    source: CalculatedDataSource
    source_feature: str
    unit: str

    @property
    def source_features(self) -> list[str]:
        return self.source_feature.split(",")

    @property
    def column_key(self) -> str:
        return self.column if isinstance(self.column, str) else ",".join(self.column)


CALCULATED_DATA_REFERENCE: dict[AggregationType, list[CalculatedDataReference]] = {
    AggregationType.AVERAGE: [
        CalculatedDataReference(
            column="Total",
            name="Average Total",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count",
            unit="#",
        ),
        CalculatedDataReference(
            column=CONCENTRATION_COLUMNS,
            name="Average Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataReference(
            column="SD",
            name="SD Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataReference(
            column="Positives",
            name="Average Positives",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="positive partition count",
            unit="#",
        ),
        CalculatedDataReference(
            column="CV%",
            name="CV",
            source=CalculatedDataSource.CALCULATED_DATA,
            source_feature="Average Concentration,SD Concentration",
            unit="%",
        ),
        CalculatedDataReference(
            column="95%DeltaLCI",
            name="Average 95%DeltaLCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="95%DeltaLCI",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="95%DeltaUCI",
            name="Average 95%DeltaUCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="95%DeltaUCI",
            unit="(unitless)",
        ),
    ],
    AggregationType.POOLED: [
        CalculatedDataReference(
            column="Total",
            name="Pooled Total",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count",
            unit="#",
        ),
        CalculatedDataReference(
            column="Conc. cp/uL",
            name="Pooled Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataReference(
            column="Positives",
            name="Pooled Positives",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="positive partition count",
            unit="#",
        ),
    ],
    AggregationType.INDIVIDUAL: [
        CalculatedDataReference(
            column="95%DeltaLCI",
            name="95%DeltaLCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="95%DeltaUCI",
            name="95%DeltaUCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Precision %",
            name="Precision %",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Lambda (cp/Rxn)",
            name="Lambda (cp/Rxn)",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Lambda_95%DeltaLCI",
            name="Lambda_95%DeltaLCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Lambda_95%DeltaUCI",
            name="Lambda_95%DeltaUCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Lambda_95%DeltaLCI",
            name="Lambda_95%DeltaLCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="Lambda_95%DeltaUCI",
            name="Lambda_95%DeltaUCI",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count,positive partition count",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="MeanQC",
            name="MeanQC",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="QC Channel",
            unit="(unitless)",
        ),
        CalculatedDataReference(
            column="CV_QC",
            name="CV_QC",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="QC Channel",
            unit="(unitless)",
        ),
    ],
}



import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv


class MSDWorkbenchReader:
    SUPPORTED_EXTENSIONS = "csv, txt"
    plate_data: pd.DataFrame
    well_plate_id: str

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        data = read_csv(named_file_contents.contents)
        data = data.where(pd.notna(data), None)
        data.index = pd.Index(data.index.to_series().ffill())
        data.index = data.index.astype(str).str.strip()
        data.columns = data.columns.astype(str).str.strip()
        first_row = str(data.iloc[0].index[0])
        if "Plate" not in first_row:
            msg = "Plate ID not found in the first row of the data"
            raise AllotropeConversionError(msg)
        self.well_plate_id = first_row.split("_")[-1].strip()
        # Set the first row as the header
        data.columns = pd.Index(data.iloc[0])
        data = data[1:].reset_index(drop=True)
        self.plate_data = data



from io import StringIO

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import LinesReader, read_to_lines
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData


class ThermoFisherGenesys30Reader:
    SUPPORTED_EXTENSIONS = "csv,tsv"
    header: SeriesData
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = LinesReader(read_to_lines(named_file_contents))

        if named_file_contents.extension == "csv":
            metadata_list = [line for line in reader.pop_until("^,,") if line]
            reader.drop_until_inclusive(",,")
            delimiter = ","
        else:
            metadata_list = list(reader.pop_until_empty())
            reader.drop_until_inclusive("")
            delimiter = "\t"

        rawdata_list = list(reader.pop_until_empty())
        rawdata_string = StringIO("\n".join(rawdata_list))
        self.data = read_csv(rawdata_string, header=0, delimiter=delimiter)
        self.data.columns = self.data.columns.astype(str).str.strip()

        metadata_string = StringIO("\n".join(metadata_list))
        metadata_dataframe = (
            read_csv(
                metadata_string,
                header=None,
                delimiter=delimiter,
                keep_default_na=False,
                index_col=0,
            )
            .astype(str)
            .T
        )
        metadata_dataframe.columns = metadata_dataframe.columns.str.strip()
        metadata_dataframe = metadata_dataframe.head(1)

        self.header = df_to_series_data(metadata_dataframe)



from dataclasses import dataclass

from allotropy.allotrope.models.adm.light_obscuration.benchling._2023._12.light_obscuration import (
    CalculatedDataDocumentItem,
    DataProcessingDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    DistributionAggregateDocument,
    DistributionDocumentItem,
    DistributionItem,
    LightObscurationAggregateDocument,
    LightObscurationDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    TCalculatedDataAggregateDocument,
    TDataSourceAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCountsPerMilliliter,
    TQuantityValueMicrometer,
    TQuantityValueMilliliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TQuantityValue,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class ProcessedDataFeature:
    identifier: str
    particle_size: float
    cumulative_count: float
    cumulative_particle_density: float
    differential_particle_density: float | None = None
    differential_count: float | None = None


@dataclass(frozen=True)
class ProcessedData:
    dilution_factor_setting: float
    distributions: list[ProcessedDataFeature]


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    measurement_time: str
    sample_identifier: str
    flush_volume_setting: float
    detector_view_volume: float
    repetition_setting: int
    sample_volume_setting: float

    # Processed data
    processed_data: ProcessedData

    # Optional processed data
    calculated_data: list[CalculatedDataItem] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    analyst: str
    measurements: list[Measurement]


@dataclass(frozen=True)
class Metadata:
    file_name: str
    unc_path: str
    equipment_serial_number: str
    detector_identifier: str
    detector_model_number: str
    software_version: str
    software_name: str


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/light-obscuration/BENCHLING/2023/12/light-obscuration.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            light_obscuration_aggregate_document=LightObscurationAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    device_document=[
                        DeviceDocumentItem(
                            detector_identifier=data.metadata.detector_identifier,
                            detector_model_number=data.metadata.detector_model_number,
                        )
                    ],
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    software_name=data.metadata.software_name,
                    UNC_path=data.metadata.unc_path,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                light_obscuration_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> LightObscurationDocumentItem:
        return LightObscurationDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                analyst=measurement_group.analyst,
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, _: Metadata
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            sample_document=self._get_sample_document(measurement),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        flush_volume_setting=TQuantityValueMilliliter(
                            value=measurement.flush_volume_setting
                        ),
                        detector_view_volume=TQuantityValueMilliliter(
                            value=measurement.detector_view_volume
                        ),
                        repetition_setting=measurement.repetition_setting,
                        sample_volume_setting=TQuantityValueMilliliter(
                            value=measurement.sample_volume_setting,
                        ),
                    )
                ]
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData
    ) -> ProcessedDataAggregateDocument:
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    data_processing_document=DataProcessingDocument(
                        dilution_factor_setting=TQuantityValueUnitless(
                            value=data.dilution_factor_setting,
                        )
                    ),
                    distribution_aggregate_document=DistributionAggregateDocument(
                        distribution_document=[
                            DistributionDocumentItem(
                                distribution=[
                                    DistributionItem(
                                        distribution_identifier=feature.identifier,
                                        particle_size=TQuantityValueMicrometer(
                                            value=feature.particle_size
                                        ),
                                        cumulative_count=TQuantityValueUnitless(
                                            value=feature.cumulative_count
                                        ),
                                        cumulative_particle_density=TQuantityValueCountsPerMilliliter(
                                            value=feature.cumulative_particle_density
                                        ),
                                        differential_particle_density=quantity_or_none(
                                            TQuantityValueCountsPerMilliliter,
                                            feature.differential_particle_density,
                                        ),
                                        differential_count=quantity_or_none(
                                            TQuantityValueUnitless,
                                            feature.differential_count,
                                        ),
                                    )
                                    for feature in data.distributions
                                ]
                            )
                        ]
                    ),
                )
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> TCalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    # TODO(nstender): figure out how to convince typing that unit is correct.
                    calculated_result=TQuantityValue(  # type:ignore[arg-type]
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=TDataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )



from __future__ import annotations

from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.biorad_bioplex_manager.biorad_bioplex_manager_reader import (
    BioradBioplexReader,
)
from allotropy.parsers.biorad_bioplex_manager.biorad_bioplex_manager_structure import (
    create_measurement_group,
    create_metadata,
    SampleMetadata,
    SystemMetadata,
    Well,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class BioradBioplexParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Bio-Rad Bio-Plex Manager"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "xml"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BioradBioplexReader(named_file_contents)

        samples_dict = SampleMetadata.create_samples(reader["Samples"])
        # Only parse system-level metadata once, from the first well.
        system_metadata = SystemMetadata.create(reader["Wells"][0])

        return Data(
            create_metadata(
                reader.root, system_metadata, named_file_contents.original_file_path
            ),
            [
                create_measurement_group(
                    well,
                    samples_dict[well.name],
                    system_metadata,
                    experimental_data_id=reader["NativeDocumentLocation"].text,
                    experiment_type=reader["Description"].text,
                    plate_well_count=int(
                        reader.get_attribute("PlateDimensions", "TotalWells")
                    ),
                )
                for well in [Well.create(well_xml) for well_xml in reader["Wells"]]
            ],
        )



from __future__ import annotations

from typing import Any

import jsonschema

from allotropy.allotrope.converter import unstructure
from allotropy.allotrope.schemas import get_schema_from_model
from allotropy.exceptions import (
    AllotropeSerializationError,
    AllotropeValidationError,
)


def serialize_and_validate_allotrope(model: Any) -> dict[str, Any]:
    try:
        allotrope_dict = unstructure(model)
    except Exception as e:
        msg = f"Failed to serialize allotrope model: {e}"
        raise AllotropeSerializationError(msg) from e

    try:
        allotrope_schema = get_schema_from_model(model)
    except Exception as e:
        msg = f"Failed to retrieve schema for model: {e}"
        raise AllotropeSerializationError(msg) from e

    try:
        jsonschema.validate(
            allotrope_dict,
            allotrope_schema,
            cls=jsonschema.validators.Draft202012Validator,
        )
    except Exception as e:
        msg = f"Failed to validate allotrope model against schema: {e}"
        raise AllotropeValidationError(msg) from e
    return allotrope_dict



# generated by datamodel-codegen:
#   filename:  electrophoresis.schema.json
#   timestamp: 2024-10-15T14:18:09+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueKiloDalton,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliAbsorbanceUnitTimesMilliliter,
    TQuantityValueMilliAbsorbanceUnitTimesSecond,
    TQuantityValueMilliliter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePercent,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TBooleanValue,
    TClass,
    TDatacube,
    TDatacubeData,
    TDatacubeStructure,
    TDateTimeStampValue,
    TDateTimeValue,
    TDoubleValue,
    TIntegerValue,
    TQuantityValue,
    TStringValue,
    TUnit,
)


@dataclass(kw_only=True)
class AdmCoreREC202409ManifestSchema:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class CustomInformationDocumentItem:
    scalar_double_datum: TDoubleValue | None = None
    unit: TUnit | None = None
    scalar_string_datum: TStringValue | None = None
    scalar_timestamp_datum: TDateTimeValue | None = None
    scalar_boolean_datum: TBooleanValue | None = None
    datum_label: TStringValue | None = None


@dataclass(kw_only=True)
class ElectronicProjectRecord:
    written_name: TStringValue
    description: Any | None = None
    start_time: TDateTimeValue | None = None


@dataclass(kw_only=True)
class ErrorDocumentItem:
    error: TStringValue
    error_feature: TStringValue | None = None


@dataclass(kw_only=True)
class ErrorAggregateDocument:
    error_document: list[ErrorDocumentItem] | None = None


@dataclass(kw_only=True)
class ImageDocumentItem:
    experimental_data_identifier: TStringValue | None = None
    index: TIntegerValue | None = None


@dataclass(kw_only=True)
class ImageAggregateDocument:
    image_document: list[ImageDocumentItem] | None = None


@dataclass(kw_only=True)
class StatisticsDocumentItem:
    statistical_feature: TClass


@dataclass(kw_only=True)
class StatisticsAggregateDocument:
    statistics_document: list[StatisticsDocumentItem] | None = None


@dataclass(kw_only=True)
class AnalysisSequenceDocument:
    written_name: TStringValue
    end_time: TDateTimeValue | None = None
    file_name: TStringValue | None = None
    identifier: TStringValue | None = None
    method_identifier: TStringValue | None = None
    method_name: TStringValue | None = None
    start_time: TDateTimeValue | None = None
    UNC_path: TStringValue | None = None
    version_number: TStringValue | None = None


@dataclass(kw_only=True)
class CustomInformationAggregateDocument:
    custom_information_document: list[CustomInformationDocumentItem]


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class ElectronicSignatureDocumentItem:
    account_identifier: TStringValue
    personal_name: TStringValue
    signature_role_type: TStringValue
    time: TStringValue
    identifier: TStringValue | None = None
    measurement_identifier: TStringValue | None = None
    method_identifier: TStringValue | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ElectronicSignatureAggregateDocument:
    electronic_signature_document: list[ElectronicSignatureDocumentItem] | None = None


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem] | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    electronic_project_record: ElectronicProjectRecord | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    ASM_file_identifier: TStringValue
    data_system_instance_identifier: TStringValue
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None
    database_primary_key: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    brand_name: TStringValue | None = None
    device_identifier: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    written_name: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue | None = None
    brand_name: TStringValue | None = None
    description: Any | None = None
    device_document: list[DeviceDocumentItem] | None = None
    device_identifier: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    brand_name: TStringValue | None = None
    detection_type: TStringValue | None = None
    device_identifier: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    field_index: int | None = None
    detector_wavelength_setting: TQuantityValueNanometer | None = None
    detector_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_wavelength_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_reference_bandwidth_setting: TQuantityValueNanometer | None = (
        None
    )
    total_measurement_time_setting: TQuantityValueSecondTime | None = None
    read_interval_setting: TQuantityValueSecondTime | None = None
    number_of_scans_setting: TQuantityValueNumber | None = None
    electronic_absorbance_reference_wavelength_setting: TQuantityValueNanometer | None = (
        None
    )
    excitation_wavelength_setting: TQuantityValueNanometer | None = None
    excitation_bandwidth_setting: TQuantityValueNanometer | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    compartment_temperature: TQuantityValueDegreeCelsius | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None


@dataclass(kw_only=True)
class Peak:
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    peak_end: TQuantityValueKiloDalton | TQuantityValueMilliliter | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_position: TQuantityValueKiloDalton | TQuantityValueMilliliter | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_name: TStringValue | None = None
    identifier: TStringValue | None = None
    relative_peak_height: TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValuePercent | TQuantityValueSecondTime | None = (
        None
    )
    written_name: TStringValue | None = None
    peak_height: TQuantityValue | TQuantityValueMilliAbsorbanceUnit | TQuantityValueRelativeFluorescenceUnit | None = (
        None
    )
    capacity_factor__chromatography_: TQuantityValueUnitless | None = None
    peak_area: TQuantityValue | TQuantityValueMilliAbsorbanceUnitTimesMilliliter | TQuantityValueMilliAbsorbanceUnitTimesSecond | None = (
        None
    )
    relative_peak_area: TQuantityValuePercent | None = None
    relative_corrected_peak_area: TQuantityValuePercent | None = None
    comment: TStringValue | None = None
    retention_time: TQuantityValueSecondTime | None = None
    retention_volume: TQuantityValueMilliliter | None = None
    peak_start: TQuantityValueKiloDalton | TQuantityValueMilliliter | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    peak_selectivity__chromatography_: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution: TQuantityValueUnitless | None = None
    field_index: int | None = None
    chromatographic_peak_resolution_using_baseline_peak_widths: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_peak_width_at_half_height__JP14_: TQuantityValueUnitless | None = (
        None
    )
    peak_width_at_4_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_13_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_32_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_60_7___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_half_height: TQuantityValueSecondTime | None = None
    peak_width_at_5___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_baseline: TQuantityValueSecondTime | None = None
    peak_width_at_inflection: TQuantityValueSecondTime | None = None
    peak_width_at_10___of_height: TQuantityValueSecondTime | None = None
    peak_width: TQuantityValueSecondTime | None = None
    statistical_skew__chromatography_: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_5___height: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_10___height: TQuantityValueUnitless | None = None
    asymmetry_factor_squared_measured_at_10___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_squared_measured_at_4_4___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_measured_at_4_4___height: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_baseline: TQuantityValueUnitless | None = None
    chromatographic_peak_asymmetry_factor: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution_using_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    chromatographic_peak_resolution_using_statistical_moments: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates__chromatography_: TQuantityValueUnitless | None = None
    number_of_theoretical_plates_measured_at_60_7___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_32_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_13_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_4_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_tangent_method: TQuantityValueUnitless | None = None


@dataclass(kw_only=True)
class PeakList:
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    peak: list[Peak] | None = None


@dataclass(kw_only=True)
class DataRegionDocumentItem:
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_region_end: TQuantityValueKiloDalton | TQuantityValueMilliliter | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    data_region_identifier: TStringValue | None = None
    data_region_name: TStringValue | None = None
    data_region_area: TQuantityValue | None = None
    relative_data_region_area: TQuantityValuePercent | None = None
    comment: TStringValue | None = None
    data_region_start: TQuantityValueKiloDalton | TQuantityValueMilliliter | TQuantityValueNumber | TQuantityValueSecondTime | None = (
        None
    )
    field_index: int | None = None


@dataclass(kw_only=True)
class DataRegionAggregateDocument:
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_region_document: list[DataRegionDocumentItem] | None = None


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    batch_identifier: TStringValue | None = None
    description: Any | None = None
    location_identifier: TStringValue | None = None
    sample_role_type: TClass | None = None
    written_name: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValue
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class ElectropherogramDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_processing_document: dict[str, Any] | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None
    peak_list: PeakList | None = None
    data_region_aggregate_document: DataRegionAggregateDocument | None = None
    derived_electropherogram_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class MeasurementDocument:
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    electropherogram_data_cube: ElectropherogramDataCube | TDatacube | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    detection_type: TStringValue | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    measurement_identifier: TStringValue | None = None
    measurement_time: TDateTimeStampValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    compartment_temperature: TQuantityValueDegreeCelsius | None = None
    absorption_profile_data_cube: TDatacube | None = None
    chromatogram_data_cube: TDatacube | None = None
    three_dimensional_ultraviolet_spectrum_data_cube: TDatacube | None = None
    fluorescence_emission_profile_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocument]
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    analytical_method_identifier: TStringValue | None = None
    method_version: TStringValue | None = None
    experimental_data_identifier: TStringValue | None = None
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class ElectrophoresisDocumentItem:
    analyst: TStringValue
    measurement_aggregate_document: MeasurementAggregateDocument
    electronic_project_record: ElectronicProjectRecord | None = None
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class ElectrophoresisAggregateDocument:
    electrophoresis_document: list[ElectrophoresisDocumentItem]
    analysis_sequence_document: AnalysisSequenceDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_system_document: DataSystemDocument | None = None
    device_system_document: DeviceSystemDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    electronic_signature_aggregate_document: ElectronicSignatureAggregateDocument | None = (
        None
    )
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: AdmCoreREC202409ManifestSchema | str
    electrophoresis_aggregate_document: ElectrophoresisAggregateDocument | None = None



from __future__ import annotations

from itertools import chain
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.moldev_softmax_pro.constants import DEVICE_TYPE
from allotropy.parsers.moldev_softmax_pro.softmax_pro_structure import (
    DataElement,
    GroupBlock,
    GroupSampleData,
    PlateBlock,
    SpectrumRawPlateData,
    StructureData,
)
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(file_path: str) -> Metadata:
    return Metadata(
        asm_file_identifier=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        model_number=NOT_APPLICABLE,
        data_system_instance_id=NOT_APPLICABLE,
        software_name="SoftMax Pro",
        file_name=Path(file_path).name,
        unc_path=file_path,
    )


def _get_data_cube(
    plate_block: PlateBlock, data_element: DataElement
) -> DataCube | None:

    if plate_block.measurement_type not in (
        MeasurementType.FLUORESCENCE_CUBE_DETECTOR,
        MeasurementType.LUMINESCENCE_CUBE_DETECTOR,
        MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR,
    ):
        return None

    return DataCube(
        label=plate_block.measurement_type.value,
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double, concept="elapsed time", unit="s"
            )
        ],
        structure_measures=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept=plate_block.header.concept,
                unit=plate_block.header.unit,
            )
        ],
        dimensions=[data_element.elapsed_time],
        measures=[data_element.kinetic_measures],
    )


def _create_measurements(plate_block: PlateBlock, position: str) -> list[Measurement]:

    measurement_type = plate_block.measurement_type

    return [
        Measurement(
            type_=measurement_type,
            identifier=data_element.uuid,
            absorbance=(
                data_element.value
                if measurement_type == MeasurementType.ULTRAVIOLET_ABSORBANCE
                else None
            ),
            fluorescence=(
                data_element.value
                if measurement_type == MeasurementType.FLUORESCENCE
                else None
            ),
            luminescence=(
                data_element.value
                if measurement_type == MeasurementType.LUMINESCENCE
                else None
            ),
            profile_data_cube=_get_data_cube(plate_block, data_element),
            # A temperature of 0 indicates the temperature was not actualy read.
            compartment_temperature=data_element.temperature or None,
            # Sample document
            location_identifier=data_element.position,
            well_plate_identifier=plate_block.header.name,
            sample_identifier=data_element.sample_identifier,
            # Device Control document
            device_type=DEVICE_TYPE,
            detection_type=plate_block.header.read_mode,
            scan_position_setting=plate_block.header.scan_position,
            detector_wavelength_setting=data_element.wavelength,
            excitation_wavelength_setting=(
                plate_block.header.excitation_wavelengths[idx]
                if plate_block.header.excitation_wavelengths
                else None
            ),
            wavelength_filter_cutoff_setting=(
                plate_block.header.cutoff_filters[idx]
                if plate_block.header.cutoff_filters
                else None
            ),
            number_of_averages=plate_block.header.reads_per_well,
            detector_gain_setting=plate_block.header.pmt_gain,
            total_measurement_time_setting=plate_block.header.read_time,
            read_interval_setting=plate_block.header.read_interval,
            number_of_scans_setting=plate_block.header.kinetic_points,
            # Error documents
            error_document=data_element.error_document,
        )
        for idx, data_element in enumerate(plate_block.iter_data_elements(position))
    ]


def _create_measurement_group(
    plate_block: PlateBlock, position: str
) -> MeasurementGroup | None:

    if not (measurements := _create_measurements(plate_block, position)):
        return None

    maximum_wavelength_signal = None
    if isinstance(plate_block.block_data.raw_data, SpectrumRawPlateData):
        maximum_wavelength_signal = (
            plate_block.block_data.raw_data.maximum_wavelength_signal[position]
        )

    return MeasurementGroup(
        measurements=measurements,
        plate_well_count=plate_block.header.num_wells,
        measurement_time=DEFAULT_EPOCH_TIMESTAMP,
        maximum_wavelength_signal=maximum_wavelength_signal,
    )


def create_measurement_groups(data: StructureData) -> list[MeasurementGroup]:
    measurement_groups = [
        measurement_group
        for plate_block in data.block_list.plate_blocks.values()
        for position in plate_block.iter_wells()
        if (measurement_group := _create_measurement_group(plate_block, position))
    ]
    if not measurement_groups:
        msg = "Invalid data - the file contains invalid or missing measurement data. Unable to construct ASM."
        raise AllotropeConversionError(msg)

    return measurement_groups


def create_calculated_data(data: StructureData) -> list[CalculatedDataItem]:
    return _get_reduced_calc_docs(data) + _get_group_calc_docs(data)


def _get_calc_docs_data_sources(
    plate_block: PlateBlock, position: str
) -> list[DataSource]:
    return [
        DataSource(
            identifier=data_source.uuid,
            feature=plate_block.header.read_mode,
        )
        for data_source in plate_block.iter_data_elements(position)
    ]


def _build_calc_doc(
    name: str,
    value: float,
    data_sources: list[DataSource],
    description: str | None = None,
) -> CalculatedDataItem:
    return CalculatedDataItem(
        identifier=random_uuid_str(),
        name=name,
        value=value,
        unit=UNITLESS,
        data_sources=data_sources,
        description=description,
    )


def _get_reduced_calc_docs(data: StructureData) -> list[CalculatedDataItem]:
    return [
        _build_calc_doc(
            name="Reduced",
            value=reduced_data_element.value,
            data_sources=_get_calc_docs_data_sources(
                plate_block,
                reduced_data_element.position,
            ),
        )
        for plate_block in data.block_list.plate_blocks.values()
        for reduced_data_element in plate_block.iter_reduced_data()
    ]


def _get_group_simple_calc_docs(
    data: StructureData,
    group_block: GroupBlock,
    group_sample_data: GroupSampleData,
) -> list[CalculatedDataItem]:
    calculated_documents = []
    for group_data_element in group_sample_data.data_elements:
        data_sources = list(
            chain.from_iterable(
                _get_calc_docs_data_sources(
                    data.block_list.plate_blocks[group_data_element.plate],
                    position,
                )
                for position in group_data_element.positions
            )
        )
        for entry in group_data_element.entries:
            calculated_documents.append(
                _build_calc_doc(
                    name=entry.name,
                    value=entry.value,
                    data_sources=data_sources,
                    description=group_block.group_columns.data.get(entry.name),
                )
            )
    return calculated_documents


def _get_group_calc_docs(data: StructureData) -> list[CalculatedDataItem]:
    calculated_documents = []
    for group_block in data.block_list.group_blocks:
        for group_sample_data in group_block.group_data.sample_data:
            calculated_documents += _get_group_simple_calc_docs(
                data, group_block, group_sample_data
            )
    return calculated_documents



from dataclasses import fields, MISSING

# We must import this to get all subclasses of TQuantityValue in scope
import allotropy.allotrope.models.shared.definitions.custom  # noqa: F401
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue


def _clean_unit(unit: str) -> str:
    return unit.replace("µ", "u").replace("μ", "u")


def _make_unit_entry(quantity_value: type[TQuantityValue]) -> str | None:
    unit_field = next(field for field in fields(quantity_value) if field.name == "unit")
    if unit_field.default == MISSING:
        return None
    return _clean_unit(str(unit_field.default))


UNIT_TO_PROPERTY = {
    unit_entry: quantity_value
    for unit_entry, quantity_value in [
        (_make_unit_entry(quantity_value), quantity_value)
        for quantity_value in TQuantityValue.__subclasses__()
    ]
    if unit_entry
}

# Some input files have units with incorrect capitalization. Try to handle this by checking with case insensitive.
# We don't do this by default, however, because there are some cases where capitalization matters, e.g.
# namometer (nm) != nanomolar (nM).
LOWER_UNIT_TO_PROPERTY = {
    unit.lower(): quantity_value for unit, quantity_value in UNIT_TO_PROPERTY.items()
}


def get_quantity_class(
    unit: str | None, default: type[TQuantityValue] | None = None
) -> type[TQuantityValue] | None:
    clean_unit = _clean_unit(unit or "")
    # Try to get case-sensitive unit, falling back to default if provided, finally checking case-insensitive unit,
    # which may not always have the correct result.
    return (
        UNIT_TO_PROPERTY.get(clean_unit)
        or default
        or LOWER_UNIT_TO_PROPERTY.get(clean_unit.lower())
    )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any

from allotropy.exceptions import AllotropeConversionError


@dataclass(kw_only=True)
class TBooleanValueItem:
    field_type: TClass
    value: bool


@dataclass(kw_only=True)
class TIntValueItem:
    value: int
    field_type: str


TIntValue = int | TIntValueItem
TIntegerValue = TIntValue


@dataclass(kw_only=True)
class TDoubleValueItem:
    value: float
    field_type: str


TDoubleValue = float | TDoubleValueItem
TFloatValue = TDoubleValue
TDecimalValue = TDoubleValue


@dataclass(kw_only=True)
class TDateTimeValueItem:
    field_type: TClass
    value: str


@dataclass(kw_only=True)
class TStringValueItem:
    value: str
    field_type: str


TBooleanArray = list[bool]
TBooleanOrNullArray = list[bool | None]
TBooleanValue = bool | TBooleanValueItem
TClass = str
TDateTimeValue = str | TDateTimeValueItem
TDateTimeStampValue = TDateTimeValue
TNumberArray = list[float]
TNumberOrNullArray = list[float | None]
TStringArray = list[str]
TStringOrNullArray = list[str | None]
TStringValue = str | TStringValueItem
TTupleData = list[float | bool | str | None]
TUnit = str


class TStatisticDatumRole(Enum):
    arithmetic_mean_role = "arithmetic mean role"
    median_role = "median role"
    relative_standard_deviation_role = "relative standard deviation role"
    skewness_role = "skewness role"
    standard_deviation_role = "standard deviation role"
    variance_role = "variance role"
    maximum_value_role = "maximum value role"
    minimum_value_role = "minimum value role"


class InvalidJsonFloat(Enum):
    """Enum to represent valid float values that are invalid in JSON.

    Values were automatically generated by schema generation code."""

    NaN = "NaN"
    field_Infinity = "+Infinity"
    field_Infinity_1 = "-Infinity"

    def __truediv__(self, other: Any) -> InvalidJsonFloat:
        # The logic here is that: NaN/Any = NaN, +Infinity/Any = +Infinity, -Infinity/Any = -Infinity
        return self


# For convenience
JsonFloat = float | InvalidJsonFloat
NaN = InvalidJsonFloat.NaN


@dataclass(frozen=True, kw_only=True)
class TQuantityValue:
    value: JsonFloat
    unit: TUnit
    has_statistic_datum_role: TStatisticDatumRole | None = None
    field_type: TClass | None = None


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValue:
    value: float | None
    unit: TUnit
    has_statistic_datum_role: TStatisticDatumRole | None = None
    field_type: TClass | None = None


class FieldComponentDatatype(Enum):
    double = "double"
    float = "float"
    decimal = "decimal"
    integer = "integer"
    byte = "byte"
    int = "int"
    short = "short"
    long = "long"
    string = "string"
    boolean = "boolean"
    dateTime = "dateTime"


class Scale(Enum):
    nominal = "nominal"
    ordinal = "ordinal"
    cardinal = "cardinal"
    interval = "interval"
    range = "range"


class Type(Enum):
    linear = "linear"
    logarithmic = "logarithmic"


@dataclass(kw_only=True)
class TFunction:
    type: Type | None = Type.linear
    start: float | None = 1
    length: float | None = None
    incr: float | None = 1


@dataclass(kw_only=True)
class TDatacubeComponent:
    field_componentDatatype: FieldComponentDatatype
    concept: TClass
    unit: TUnit | None = None
    scale: Scale | None = None
    field_asm_fill_value: str | float | int | bool | None = None


TDimensionArray = TNumberArray | TBooleanArray | TStringArray


TMeasureArray = TNumberOrNullArray | TBooleanOrNullArray | TStringOrNullArray


@dataclass(kw_only=True)
class TDatacubeStructure:
    dimensions: list[TDatacubeComponent]
    measures: list[TDatacubeComponent]


"""
TODO: datamodel-codegen cannot properly generate the models for TDatacubeData with the
oneOf{measures, points} constraint. I can't figure out how to do it correctly right now either.
We need to either figure this out and file a bug with datamodel-codegen or fix the schema.

@dataclass(kw_only=True)
class TDimensionData:
    dimensions: list[TDimensionArray | TFunction]


@dataclass(kw_only=True)
class TMeasureDatum:
    measures: list[TMeasureArray]


@dataclass(kw_only=True)
class TMeasureDatum1:
    points: list[TTupleData]


TMeasureData = TMeasureDatum | TMeasureDatum1


@dataclass(kw_only=True)
class TDatacubeData(TDimensionData):
    pass
"""


@dataclass(kw_only=True)
class TDatacubeData:
    dimensions: list[TDimensionArray | TFunction]
    measures: list[TMeasureArray] | None = None
    points: list[TTupleData] | None = None

    def __post_init__(self) -> None:
        # Logic for enforcing oneOf
        if not (self.measures is None) ^ (self.points is None):
            msg = "Exactly one of measures or points must be set on a datacube."
            raise AllotropeConversionError(msg)


@dataclass(kw_only=True)
class TDatacube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class OrderedItem:
    field_index: int | None = None



from __future__ import annotations

from re import search

from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
    Peak,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.creator import (
    create_data_cube,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.transformations import (
    Transformation,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)
from allotropy.parsers.utils.uuids import random_uuid_str


class UnicornMeasurement(Measurement):
    @classmethod
    def filter_curve_or_none(
        cls, curve_elements: list[StrictXmlElement], pattern: str
    ) -> StrictXmlElement | None:
        for element in curve_elements:
            name_element = element.find("Name")
            if name := name_element.get_text_or_none():
                if search(pattern, name):
                    return element
        return None

    @classmethod
    def get_data_cube_handler_or_none(
        cls, handler: UnicornZipHandler, curve_element: StrictXmlElement
    ) -> UnicornZipHandler | None:
        names = ["CurvePoints", "CurvePoint", "BinaryCurvePointsFileName"]
        if data_name := curve_element.recursive_find_or_none(names):
            if name := data_name.get_text_or_none():
                return handler.get_zip_from_pattern(name)
        return None

    @classmethod
    def get_data_cube_or_none(
        cls,
        handler: UnicornZipHandler,
        curve: StrictXmlElement | None,
        data_cube_component: DataCubeComponent,
        transformation: Transformation | None = None,
    ) -> DataCube | None:
        if curve is None:
            return None

        if data_cube_handler := cls.get_data_cube_handler_or_none(handler, curve):
            name_element = curve.find("Name")
            if name := name_element.get_text_or_none():
                return create_data_cube(
                    data_cube_handler,
                    name,
                    data_cube_component,
                    transformation,
                )
        return None

    @classmethod
    def get_measurement(
        cls,
        static_docs: StaticDocs,
        device_control_docs: list[DeviceControlDoc],
        chromatogram_data_cube: DataCube | None = None,
        processed_data_chromatogram_data_cube: DataCube | None = None,
        derived_column_pressure_data_cube: DataCube | None = None,
        peaks: list[Peak] | None = None,
    ) -> UnicornMeasurement:
        return UnicornMeasurement(
            measurement_identifier=random_uuid_str(),
            chromatography_serial_num=static_docs.chromatography_serial_num,
            column_inner_diameter=static_docs.column_inner_diameter,
            chromatography_chemistry_type=static_docs.chromatography_chemistry_type,
            chromatography_particle_size=static_docs.chromatography_particle_size,
            injection_identifier=static_docs.injection_identifier,
            injection_time=static_docs.injection_time,
            autosampler_injection_volume_setting=static_docs.autosampler_injection_volume_setting,
            sample_identifier=static_docs.sample_identifier,
            batch_identifier=static_docs.batch_identifier,
            chromatogram_data_cube=chromatogram_data_cube,
            device_control_docs=device_control_docs,
            processed_data_chromatogram_data_cube=processed_data_chromatogram_data_cube,
            derived_column_pressure_data_cube=derived_column_pressure_data_cube,
            sample_custom_info={
                "sample_identifier_2": static_docs.sample_identifier_2,
                "sample_identifier_3": static_docs.sample_identifier_3,
                "sample_volume_2": static_docs.sample_volume_2,
                "sample_volume_3": static_docs.sample_volume_3,
            },
            peaks=peaks,
        )

    @classmethod
    def is_valid(cls, data_cubes: list[DataCube | None]) -> bool:
        return any(data_cube is not None for data_cube in data_cubes)



from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import (
    ContainerType,
    SampleRoleType,
)

SAMPLE_ROLE_TYPES_MAP = {
    "Unkn": SampleRoleType.unknown_sample_role,
    "Std": SampleRoleType.standard_sample_role,
    "NTC": SampleRoleType.control_sample_role,
    "Pos": SampleRoleType.control_sample_role,
    "Neg": SampleRoleType.control_sample_role,
    "NRT": SampleRoleType.control_sample_role,
}

DISPLAY_NAME = "Bio-Rad CFX Maestro"
DEVICE_TYPE = "qPCR"

PRODUCT_MANUFACTURER = "Bio-Rad"
SOFTWARE_NAME = "CFX Maestro"
CONTAINER_TYPE = ContainerType.well_plate
EXPERIMENT_TYPE = "comparative CT qPCR experiment"



from io import StringIO

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.constants import (
    MULTIPLATE_FILE_ERROR,
    NO_PLATE_DATA_ERROR,
)
from allotropy.parsers.lines_reader import SectionLinesReader
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class AgilentGen5Reader:
    header_data: SeriesData
    sections: dict[str, list[str]]
    SUPPORTED_EXTENSIONS = "txt"

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = SectionLinesReader.create(named_file_contents)

        plate_readers = list(reader.iter_sections("^Software Version"))
        if not plate_readers:
            raise AllotropeConversionError(NO_PLATE_DATA_ERROR)

        if len(plate_readers) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        plate_reader = plate_readers[0]

        # Header data is all data until the "Procedure Details" section, read in as a cvs
        assert_not_none(
            plate_reader.drop_until("^Software Version"), "Software Version"
        )
        lines = [line for line in plate_reader.pop_until("Procedure Details") if line]
        df = read_csv(
            StringIO("\n".join(lines)),
            header=None,
            index_col=0,
            keep_default_na=False,
            sep="\t",
        ).T
        self.header_data = df_to_series_data(df)

        self.sections = {}

        # Special handling for the "Procedure Details" section, which has an empty line after the title.
        assert_not_none(
            plate_reader.drop_until("^Procedure Details"), "Procedure Details"
        )
        plate_reader.pop()
        plate_reader.drop_empty()
        # add the plate type to the header data to extract the well count if needed
        self.header_data.series["Plate Type"] = plate_reader.get()
        self.sections["Procedure Details"] = list(plate_reader.pop_until_empty())
        plate_reader.drop_empty()

        while plate_reader.current_line_exists():
            lines = list(plate_reader.pop_until_empty())
            self.sections[lines[0].split("\t")[0].strip(":")] = lines
            plate_reader.drop_empty()



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_skanit.thermo_skanit_structure import DataThermoSkanIt
from allotropy.parsers.utils.pandas import read_multisheet_excel
from allotropy.parsers.vendor_parser import VendorParser


class ThermoSkanItParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Thermo Fisher Scientific SkanIt"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "xlsx"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        contents = read_multisheet_excel(
            named_file_contents.contents, engine="calamine"
        )
        return DataThermoSkanIt.create(
            sheet_data=contents, file_path=named_file_contents.original_file_path
        )



from __future__ import annotations

from io import StringIO
import re

import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import LinesReader, SectionLinesReader
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    parse_header_row,
    read_csv,
    read_multisheet_excel,
    SeriesData,
    split_dataframe,
    split_header_and_data,
)
from allotropy.parsers.utils.values import assert_not_none


class AppBioQuantStudioReader:
    SUPPORTED_EXTENSIONS = "txt,xlsx"
    header: SeriesData
    sections: dict[str, pd.DataFrame]

    @staticmethod
    def create(named_file_contents: NamedFileContents) -> AppBioQuantStudioReader:
        if named_file_contents.extension == "xlsx":
            return AppBioQuantStudioXLSXReader(named_file_contents)
        else:
            return AppBioQuantStudioTXTReader(named_file_contents)


class AppBioQuantStudioXLSXReader(AppBioQuantStudioReader):
    def __init__(self, named_file_contents: NamedFileContents) -> None:
        raw_contents = read_multisheet_excel(
            named_file_contents.contents,
            header=None,
            engine="calamine",
        )
        contents = {name: df.replace(np.nan, None) for name, df in raw_contents.items()}
        self.header = self.get_header(contents)
        self.sections = self.get_sections(contents)

    def get_header(self, contents: dict[str, pd.DataFrame]) -> SeriesData:
        sheet = next(iter(contents.values()))
        df, _ = split_header_and_data(sheet, lambda row: row[0] is None)
        return df_to_series_data(parse_header_row(df.T))

    def get_sections(
        self, contents: dict[str, pd.DataFrame]
    ) -> dict[str, pd.DataFrame]:
        sections = {}
        for name, sheet in contents.items():
            _, data = split_header_and_data(sheet, lambda row: row[0] is None)
            if name == "Results":
                data = data.reset_index(drop=True)
                data, metadata = split_dataframe(data, lambda row: row[0] is None)
                if metadata is not None:
                    sections["Results Metadata"] = parse_header_row(metadata.T)

            sections[name] = parse_header_row(data.replace(np.nan, None))
        return sections


class AppBioQuantStudioTXTReader(AppBioQuantStudioReader):
    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = SectionLinesReader.create(named_file_contents)
        self.header = self.get_header(reader)
        self.sections = self.get_sections(reader)

    def get_header(self, reader: LinesReader) -> SeriesData:
        lines = [line.strip() for line in reader.pop_until(r"^\[.+\]") if line.strip()]
        if not lines:
            msg = "Cannot parse data from empty header."
            raise AllotropeConversionError(msg)

        csv_stream = StringIO("\n".join(lines))
        raw_data = read_csv(
            csv_stream, header=None, sep="=", skipinitialspace=True, index_col=0
        )
        raw_data.index = raw_data.index.str.replace("*", "")
        return df_to_series_data(raw_data.T.replace(np.nan, None))

    def get_sections(self, reader: SectionLinesReader) -> dict[str, pd.DataFrame]:
        sections = {}
        for section_reader in reader.iter_sections(r"^\[.+\]"):
            match = re.match(
                r"^\[(.+)\]",
                assert_not_none(section_reader.pop(), "Unexpected empty section"),
            )
            title = str(
                assert_not_none(
                    match, f"Cannot read title section: {section_reader.get()}"
                ).groups()[0]
            )
            data_lines = list(section_reader.pop_until_empty())
            section_reader.drop_empty()
            metadata_lines = list(section_reader.pop_until_empty())
            if title == "Results" and metadata_lines:
                # Treat results metadata as an additional section
                csv_stream = StringIO("\n".join(metadata_lines))
                sections["Results Metadata"] = read_csv(
                    csv_stream, header=None, sep="=", skipinitialspace=True, index_col=0
                ).T

            sections[title] = read_csv(
                StringIO("\n".join(data_lines)), sep="\t", thousands=r","
            ).replace(np.nan, None)
        return sections



from __future__ import annotations

from pathlib import PureWindowsPath
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.constants import NOT_APPLICABLE, round_to_nearest_well_count
from allotropy.parsers.ctl_immunospot import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
)


def _create_measurement(
    well_row: str,
    well_col: str,
    well_plate_identifier: str,
    plate_data: dict[str, pd.DataFrame],
    histograms: dict[str, tuple[list[float], list[float]]],
) -> Measurement:
    location_identifier = f"{well_row}{well_col}"
    return Measurement(
        type_=MeasurementType.OPTICAL_IMAGING,
        device_type=constants.DEVICE_TYPE,
        identifier=random_uuid_str(),
        well_plate_identifier=well_plate_identifier,
        location_identifier=location_identifier,
        sample_identifier=f"{well_plate_identifier}_{location_identifier}",
        detection_type=constants.DETECTION_TYPE,
        processed_data=ProcessedData(
            identifier=random_uuid_str(),
            features=[
                ImageFeature(
                    identifier=random_uuid_str(),
                    feature=name,
                    result=float(data[well_col][well_row]),
                )
                for name, data in plate_data.items()
            ],
        ),
        custom_data_cubes=[
            DataCube(
                label="spot count histogram",
                structure_dimensions=[
                    DataCubeComponent(
                        concept="spot size",
                        type_=FieldComponentDatatype.double,
                        unit=UNITLESS,
                    )
                ],
                structure_measures=[
                    DataCubeComponent(
                        concept="spot count",
                        type_=FieldComponentDatatype.double,
                        unit="Number",
                    )
                ],
                dimensions=[histograms[location_identifier][0]],
                measures=[histograms[location_identifier][1]],
            )
        ]
        if histograms and location_identifier in histograms
        else None,
    )


def create_measurement_groups(
    header: SeriesData,
    plate_identifier: str | None,
    plate_data: dict[str, pd.DataFrame],
    histograms: dict[str, tuple[list[float], list[float]]],
) -> list[MeasurementGroup]:
    well_plate_identifier = (
        plate_identifier or PureWindowsPath(header[str, "File path"]).stem
    )
    first_plate: pd.DataFrame = next(iter(plate_data.values()))
    plate_well_count = assert_not_none(
        round_to_nearest_well_count(first_plate.size),
        f"Unable to determine valid plate count from dataframe of size: {first_plate.size}",
    )
    return [
        MeasurementGroup(
            plate_well_count=plate_well_count,
            measurement_time=header[str, ("Counted", "Review Date")],
            analyst=header[str, "Authenticated user"],
            measurements=[
                _create_measurement(
                    row, col, well_plate_identifier, plate_data, histograms
                )
            ],
        )
        for row in first_plate.index
        for col in first_plate.columns
    ]


def create_metadata(header: SeriesData) -> Metadata:
    path = PureWindowsPath(header[str, "File path"])
    return Metadata(
        file_name=path.name,
        unc_path=str(path),
        device_identifier=NOT_APPLICABLE,
        model_number=assert_not_none(
            re.match(r"^(\w+)-(\w+)", header[str, "Analyzer Serial number"]),
            msg="Unable to parse analyzer serial number.",
        ).group(1),
        data_system_instance_id=header[str, "Computer name"],
        equipment_serial_number=header[str, "Analyzer Serial number"],
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=constants.SOFTWARE_NAME,
        software_version=assert_not_none(
            re.match(r"^ImmunoSpot ([\d\.]+)$", header[str, "Software version"]),
            msg="Unable to parse software version",
        ).group(1),
    )



import logging
from pathlib import Path

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ScanPositionSettingPlateReader,
)
from allotropy.parsers.bmg_labtech_smart_control.bmg_labtech_smart_control_reader import (
    BmgLabtechSmartControlReader,
)
from allotropy.parsers.bmg_labtech_smart_control.constants import (
    DEVICE_TYPE,
    PRODUCT_MANUFACTURER,
    SAMPLE_ROLE_TYPE_MAPPING,
    SOFTWARE_NAME,
    TARGET_TEMPERATURE,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none, try_float


def create_metadata(file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        device_identifier=NOT_APPLICABLE,
        product_manufacturer=PRODUCT_MANUFACTURER,
        model_number=NOT_APPLICABLE,
        data_system_instance_id=NOT_APPLICABLE,
        file_name=path.name,
        unc_path=file_path,
        software_name=SOFTWARE_NAME,
    )


def map_measurement_group(row: SeriesData, headers: SeriesData) -> MeasurementGroup:
    fluorescence = row[
        float,
        f"Raw Data ({headers[str, 'Excitation:']}/{headers[str, 'Emission:']})",
    ]
    sample_identifier = row[str, "Content"]
    filter_position_map = {
        "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
        "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
    }
    scan_position_setting = (
        filter_position_map.get("Bottom")
        if headers.get(str, "Top optic used") is None
        else filter_position_map.get("Top")
    )
    mass_concentration = row.get(float, "Standard Concentrations [ng/mL]")
    if mass_concentration is not None:
        # since this is ng/mL, we need to convert to pg/mL by multiplying by 1000
        mass_concentration *= 1000

    return MeasurementGroup(
        analyst=headers.get(str, "User"),
        measurement_time=f"{headers[str, 'Date']} {headers[str, 'Time']}",
        experimental_data_identifier=headers.get(str, "ID1"),
        experiment_type=headers.get(str, "Test Name"),
        plate_well_count=try_float(
            headers[str, "Microplate name:"].split()[-1], "plate well count"
        ),
        measurements=[
            Measurement(
                identifier=random_uuid_str(),
                compartment_temperature=row.get(float, TARGET_TEMPERATURE),
                fluorescence=fluorescence,
                sample_identifier=sample_identifier,
                sample_role_type=SAMPLE_ROLE_TYPE_MAPPING.get(
                    sample_identifier.split()[0]
                ),
                location_identifier=row[str, "Well"],
                well_plate_identifier=headers.get(str, "ID3"),
                mass_concentration=mass_concentration,
                device_type=DEVICE_TYPE,
                detection_type=headers[str, "Measurement type:"].split()[0].lower(),
                detector_distance_setting=headers.get(float, "Focal height  [mm]:"),
                number_of_averages=headers.get(float, "No. of flashes per well:"),
                detector_gain_setting=headers.get(str, "Gain obtained by:"),
                scan_position_setting=scan_position_setting,
                type_=MeasurementType.FLUORESCENCE,
                sample_custom_info={
                    "dilution factor setting": quantity_or_none(
                        TQuantityValueUnitless, row.get(float, "Dilutions")
                    ),
                    "container identifier": headers.get(str, "Microplate name:"),
                },
                device_control_custom_info={
                    "settling time setting": quantity_or_none(
                        TQuantityValueSecondTime,
                        headers.get(float, "Settling time [s]:"),
                    ),
                    "reading direction setting": headers.get(str, "Reading direction:"),
                    "dichroic filter setting": headers.get(str, "Dichroic filter:"),
                    "optic preset name": headers.get(str, "Presetname:"),
                    "spoon type": headers.get(str, "Spoon type:"),
                    "well used for focus adjustment": headers.get(
                        str, "Well used for focus adjustment:"
                    ),
                    "focal height obtained by": headers.get(
                        str, "Focal height obtained by:"
                    ),
                },
                detector_wavelength_setting=try_float(
                    headers[str, "Emission:"].split("-")[0], "emission"
                ),
                detector_bandwidth_setting=try_float(
                    headers[str, "Emission:"].split("-")[1], "emission"
                ),
                excitation_bandwidth_setting=try_float(
                    headers[str, "Excitation:"].split("-")[1], "excitation"
                ),
                excitation_wavelength_setting=try_float(
                    headers[str, "Excitation:"].split("-")[0], "excitation"
                ),
            )
        ],
    )


def create_calculated_data_documents(
    measurement_groups: list[MeasurementGroup], reader: BmgLabtechSmartControlReader
) -> list[CalculatedDataItem] | None:
    blank_measurements = [
        group.measurements[0]
        for group in measurement_groups
        if group.measurements[0].sample_role_type == SampleRoleType.blank_role
    ]

    if not blank_measurements:
        return []

    if not reader.average_of_blank_used:
        logging.warning(
            "Microplate endpoint sheet does not contain average of blanks used value but blank measurements are present."
        )
        return []
    non_blank_measurements = [
        measurement
        for group in measurement_groups
        for measurement in group.measurements
        if measurement.sample_role_type != SampleRoleType.blank_role
    ]

    blank_average = reader.average_of_blank_used
    average_calc_document = CalculatedDataItem(
        identifier=random_uuid_str(),
        name="Average of all blanks used",
        value=blank_average,
        data_sources=[
            DataSource(
                identifier=measurement.identifier,
                feature="fluorescence",
            )
            for measurement in blank_measurements
        ],
        unit="RFU",
    )
    blank_corrected_calc_documents: list[CalculatedDataItem] = []
    for idx, measurement in enumerate(non_blank_measurements):
        corrected_value = SeriesData(reader.data.iloc[idx])[
            float, "Blank corrected based on Raw Data (480-14/520-30)"
        ]
        blank_corrected_calc_documents.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name="Blank corrected based on Raw Data (480-14/520-30)",
                value=corrected_value,
                data_sources=[
                    DataSource(
                        identifier=measurement.identifier,
                        feature="fluorescence",
                    ),
                    DataSource(
                        identifier=average_calc_document.identifier,
                        feature="Average of all blanks used",
                    ),
                ],
                unit="RFU",
            )
        )
    return [average_calc_document, *blank_corrected_calc_documents]



PHARMSPEC_SOFTWARE_NAME = "PharmSpec"
DEVICE_TYPE = "solution-analyzer"

# This map is used to map the column names to units, for calculated data items.
UNIT_LOOKUP = {
    "particle_size": "µm",
    "cumulative_count": "(unitless)",
    "cumulative_particle_density": "Counts/mL",
    "differential_particle_density": "Counts/mL",
    "differential_count": "(unitless)",
}
VALID_CALCS = ["Average"]

REQUIRED_DISTRIBUTION_DOCUMENT_KEYS = [
    "particle_size",
    "cumulative_count",
    "cumulative_particle_density",
    "differential_particle_density",
    "differential_count",
]



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    read_csv,
    SeriesData,
)


class ChemometecNcViewReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        self.data = read_csv(named_file_contents.contents)
        self.header = df_to_series_data(self.data.head(1))
